#========================================================================================
# §package      DE_EPAGES::Amazon::API::Object::AmazonConfig
# §state        public
# §base         DE_EPAGES::ProductPortal::API::Object::AmazonConfig
# §description  object interface for AmazonConfig.
#========================================================================================
package DE_EPAGES::Amazon::API::Object::AmazonConfig;
use base DE_EPAGES::ProductPortal::API::Object::PortalConfig;

use DateTime;
use DateTime::Duration;
use LWP::UserAgent;
use HTTP::Request;
use URI ();
use URI::QueryParam;
use URI::Escape qw(uri_escape_utf8);
use Digest::SHA qw ( sha256_hex hmac_sha256 hmac_sha256_hex );
use JSON;
use DE_EPAGES::Amazon::API::Constants
  qw ( %AMAZON_ONLINESTATUS %AMAZON_ITEM_CONDITIONS %AMAZON_FEEDS %AMAZON_PROCESSINGSTATUS ItemConditionValues
      MIN_DELIVERY_LATENCY MAX_DELIVERY_LATENCY $OAUTH_TOKEN_ENDPOINT REPORT_SCHEDULER_PERIOD
       %AMAZON_PENDING_FEEDS %REPORTS_TIME_SPAN);
use DE_EPAGES::Amazon::API::Tools qw ( WantList );
use DE_EPAGES::Amazon::API::AmazonTransaction;
use DE_EPAGES::Amazon::API::RemoteAdapterFactory;
use DE_EPAGES::Amazon::API::Object::AmazonSubmitFeedQueue;
use DE_EPAGES::Core::API::Log qw ( LogDebug LogCommunication LogError );
use DE_EPAGES::Core::API::Error qw( Error ExistsError GetError );
use DE_EPAGES::Core::API::Object::NumberFormat;
use DE_EPAGES::Core::API::Warning qw( Warning );
use DE_EPAGES::Core::API::String qw ( QuoteXML );
use DE_EPAGES::Core::API::TestParameter qw ( TestInteger );
use DE_EPAGES::Core::API::PerlTools qw( GetCurrentDateTime GenerateShortGUID AddDuration );
use DE_EPAGES::Database::API::Connection qw( GetCurrentDBHandle );
use DE_EPAGES::Object::API::Factory qw ( Transaction LoadClassByAlias LoadRootObject );
use DE_EPAGES::Object::API::Table::Currency qw ( GetInfoCurrency );
use DE_EPAGES::XML::API::DateTimeFormatXML;
use strict;

use constant MAX_PRICEUPDATES_PER_CALL => 10;

sub ErrorDefinition {
    return {
        OrderCountMismatch =>
'Please report error! Imported orders (#ImportOrderCount) does not equal transactions in xml report (#XMLTransactionCount). '.
'Report number:#ReportNumber, PortalConfig:#PortalConfigAlias, Shop:#ShopAlias, Store:#Store.',
        EndOfferFailed =>
'Removing offer #AmazonOfferID failed, PortalConfig:#PortalConfigAlias, Shop:#ShopAlias, Store:#Store.',
        OfferStillOnline =>
'Warning: Removing Amazon Offer which is still online: #AmazonOfferID, PortalConfig:#PortalConfigAlias, Shop:#ShopAlias, Store:#Store.',
        PortalConfigMismatch =>
'PortalConfig (#AmazonOfferPortalConfigAlias) of AmazonOffer (#AmazonOfferAlias) is not matching PortalConfig (#PortalConfigAlias) for upload. Shop:#ShopAlias, Store:#Store.',
        ProductOfferExists =>
'Product (#ProductAlias) already exists on Amazon Platform (#PortalConfigAlias). Shop:#ShopAlias, Store:#Store.',
        OfferNoPrice =>
'Offer for product (#ProductAlias) is created without price for Amazon Platform (#PortalConfigAlias). Shop:#ShopAlias, Store:#Store.',
        OfferUPCEANMissing =>
        'UPC or EAN code missing. Cannot upload offer.',
        OfferNoCondition =>
        'Item Condition is missing. Cannot upload Offer (#AmazonOfferID). Shop:#ShopAlias, Store:#Store.',
        OrderReportMessageMismatch =>
        'OrderReport message count mismatch. Expected: #ExpectedCount, received #ReceivedCount. Feed:#FeedAlias, PortalConfig:#PortalConfigAlias, Shop:#ShopAlias, Store:#Store.',
        AmazonMerchantSettingsIncomplete =>
          'Your Amazon settings are incomplete. Please configure your Amazon account. Error: #ErrorMessage',
    };
}

#========================================================================================
# §function     $self->logToCommunication
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->logToCommunication($Name, $Value);
#----------------------------------------------------------------------------------------
# §description  Log information to Communication.log
#----------------------------------------------------------------------------------------
# §input        $Name | Name to log under | string
# §input        $Value | Value to log | *.*
#========================================================================================
sub logToCommunication {
    my $self = shift;
    my ($Name, $Value) = @_;

    my $MainPortal = $self->get('Portal');

    return if ! defined $MainPortal;
    return if ! $MainPortal->get('Logging');

    LogCommunication($Name, $Value, 'Amazon');
}

#========================================================================================
# §function     $self->signRequest
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->signRequest($Method, $Url, $Content, $hHeaders);
#----------------------------------------------------------------------------------------
# §description  Sign the request as required by Amazon for their SP-API

#                   https://developer-docs.amazon.com/sp-api/docs/connecting-to-the-selling-partner-api
#
#                   - https://docs.aws.amazon.com/general/latest/gr/sigv4-create-canonical-request.html
#                   - https://docs.aws.amazon.com/general/latest/gr/sigv4-create-string-to-sign.html
#                   - https://docs.aws.amazon.com/general/latest/gr/sigv4-calculate-signature.html
#                   - https://docs.aws.amazon.com/general/latest/gr/sigv4-add-signature-to-request.html
#----------------------------------------------------------------------------------------
# §input        $Method | HTTP method e.g. GET, POST | string
# §input        $Url | URL to connect to | string
# §input        $Content | Request body content | string
# §input        $hHeaders | Headers to add to the request | ref.hash
#========================================================================================
sub signRequest {
    my $self = shift;
    my ( $Method, $Url, $Content, $hHeaders) = @_;

    #-- First things first, add the 'x-amz-date' header
    my $Now = GetCurrentDateTime()->set_time_zone('UTC');
    my $XAmzDateHeader = $Now->ymd('') . 'T' . $Now->hms('') . 'Z';
    $hHeaders->{'x-amz-date'} = $XAmzDateHeader;

    #-- http://www.a.com/1/2/3/?A=1&A=2&B=3
    #-- |      |_______||_____| |_________|
    #-- |          |       |         |
    #-- protocol  host    path     query

    my $URI = URI->new("$Url");

    #-- This is escaped already, but we need to escape it again according to Amazon
    #-- If this is empty, it returns '/', which is what we need
    my $CanonicalURI = $URI->path;
    #-- ENCODE PATH SEGMENTS ONLY. Therefore, we need to also tell it to NOT encode '/' characters
    $CanonicalURI = uri_escape_utf8($CanonicalURI, "^A-Za-z0-9\-\._~\/");

    #-- http://www.a.com/1/2/3/?A=1&A=2&B=3 becomes:
    #-- { B => 3, A => [1, 2]}
    my $hQueryParams = $URI->query_form_hash;
    my $CanonicalQuery = '';
    #-- Double-encode any equals ( = )(%3D) characters in parameter values.
    foreach my $Key ( sort keys %$hQueryParams ) {
        if ( ref $hQueryParams->{$Key} eq 'ARRAY' ) {
            my $aValues = $hQueryParams->{$Key};
            foreach ( sort @$aValues ) {
                my $Val = $_;
                $CanonicalQuery .= uri_escape_utf8($Key) . '=' . uri_escape_utf8($Val) . '&';
            }
        } else {
            my $Val = $hQueryParams->{$Key};
            $CanonicalQuery .= uri_escape_utf8($Key) . '=' . uri_escape_utf8($Val) . '&';
        }
    }
    chop $CanonicalQuery;

    #-- Just use host/x-amz-date headers as a bare minimum, because why complicate the damn thing
    #-- Did you know that Amazon expects an empty line after this thing? But their documentation specifies otherwise? 'natch!
    my $CanonicalHeaders = 'host:' . $URI->host . "\n";
    $CanonicalHeaders .= 'x-amz-date:' . $hHeaders->{'x-amz-date'} . "\n";

    my $SignedHeaders = "host;x-amz-date";

    my $HashedPayLoad = lc sha256_hex($Content);

    my $CanonicalRequest = "$Method\n$CanonicalURI\n$CanonicalQuery\n$CanonicalHeaders\n$SignedHeaders\n$HashedPayLoad";
    #$self->logToCommunication("CanonicalRequest", $CanonicalRequest);

    my $FinalCanonicalRequest = lc sha256_hex($CanonicalRequest);

    my $PortalSite = $self->get('PortalSite');
    my $hPortalSiteInfo = $PortalSite->get(['SPAPI_Region', 'SPAPI_IAMUserAccessKeyID', 'SPAPI_IAMUserAccessKeySecret']);
    my $Region = $hPortalSiteInfo->{'SPAPI_Region'};

    #-- Important: Use UTC so that timezone is not a factor
    #-- Amazon will complain about time differences greater than 5 minutes
    my $YMD = $Now->ymd('');
    my $Service = 'execute-api';
    my $Algorithm = "AWS4-HMAC-SHA256\n";
    my $RequestDate = $hHeaders->{'x-amz-date'} . "\n"; #-- The date component should match $YMD
    my $CredentialScope = "$YMD/$Region/$Service/aws4_request\n";

    my $StringToSign = "$Algorithm$RequestDate$CredentialScope$FinalCanonicalRequest";
    #$self->logToCommunication("StringToSign", $StringToSign);

    my $KSecret = $hPortalSiteInfo->{'SPAPI_IAMUserAccessKeySecret'};
    my $KDate = hmac_sha256( $YMD, 'AWS4' . $KSecret );
    my $KRegion = hmac_sha256( $Region, $KDate );
    my $KService = hmac_sha256( $Service, $KRegion );
    my $KSigning = hmac_sha256( 'aws4_request', $KService );

    my $Signature = hmac_sha256_hex($StringToSign, $KSigning);

    my $AWSKeyID = $hPortalSiteInfo->{'SPAPI_IAMUserAccessKeyID'};
    $hHeaders->{'Authorization'} = "AWS4-HMAC-SHA256 Credential=$AWSKeyID/$CredentialScope, SignedHeaders=$SignedHeaders, Signature=$Signature";
    $hHeaders->{'x-amz-access-token'} = $self->getSPAPIAccessToken() if ! exists $hHeaders->{'x-amz-access-token'};
}

#========================================================================================
# §function     $self->doRequest
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->doRequest($Method, $Url, $Content, $hHeaders);
#----------------------------------------------------------------------------------------
# §description  Perform a HTTP request with the given parameters
#----------------------------------------------------------------------------------------
# §input        $Method | HTTP method e.g. GET, POST | string
# §input        $Url | URL to connect to | string
# §input        $Content | Request body content | string
# §input        $hHeaders | Headers to add to the request | ref.hash
# §input        $hAdditionalInfo | Additional information to log | ref.hash
#----------------------------------------------------------------------------------------
# §return       $Res | HTTP Response object | HTTP::Response
#========================================================================================
sub doRequest {
    my $self = shift;
    my ( $Method, $Url, $Content, $hHeaders, $hAdditionalInfo ) = @_;

    my $SignRequest = $hAdditionalInfo->{'SignRequest'};
    my $RDTPath = $hAdditionalInfo->{'RDTPath'};
    # If no ThrottlingWaitTime was handover from the calling method we use 1 sec sleep otherwise use the ThrottlingWaitTime
    my $ThrottlingWaitTime = $hAdditionalInfo->{'ThrottlingWaitTime'} // 1;
    #-- Requesting restricted information, requires an RDT (restricted data token)
    if ( $RDTPath ) {
        #-- Recursively calls doRequest ...
        my $RDToken = $self->getRDToken($Method, $Url, $Content, $hHeaders, $hAdditionalInfo);
        $hHeaders->{'x-amz-access-token'} = $RDToken;

        $self->signRequest($Method, $Url, $Content, $hHeaders);
    }
    elsif ( $SignRequest ) {
        #-- We need to sign this request
        $self->signRequest($Method, $Url, $Content, $hHeaders);
    }

    #-- Data structures (would appear to) get mangled by Data Dumper, so make a copy for the logging side and keep the original as it was
    my $hLogHeaders = defined $hHeaders ? {%$hHeaders} : undef;

    my $LogSuffix = GenerateShortGUID();

    my $ShopAlias = $self->getSite()->alias;
    my $StoreAlias = LoadRootObject()->get('StoreName');

    my $hReqLog = {
      Method          => $Method,
      Url             => $Url,
      Content         => $Content,
      hHeaders        => $hLogHeaders,
      hAdditionalInfo => $hAdditionalInfo,
      ShopAlias       => $ShopAlias,
      StoreAlias      => $StoreAlias,
    };
    $self->logToCommunication("Request $LogSuffix", $hReqLog);

    my $Agent = LWP::UserAgent->new();
    my $Req = HTTP::Request->new();

    $Req->method( $Method );
    $Req->uri( $Url );
    #-- Headers are case sensitive for Amazon, prefix ':' will keep the case for us
    $Req->header( ":$_" => $hHeaders->{$_} ) foreach keys %$hHeaders;

    $Req->content( $Content ) if $Content;    ##no critic qw(DE_EPAGES::Style::RequireCorrectImports)

    my $Res;
    eval {
        $Res = $Agent->request($Req);
    };
    if ( ExistsError() ) {
        my $Error = GetError();
        LogError($Error);
        return undef;
    }

    # Please see documentation Usage Plans and Rate limiting
    # https://developer-docs.amazon.com/sp-api/docs/usage-plans-and-rate-limits-in-the-sp-api
    if ($Res->code == 429) {
        $self->logToCommunication("QuotaExceeded $LogSuffix", $hReqLog);
        sleep $ThrottlingWaitTime;
        return $self->doRequest($Method, $Url, $Content, $hHeaders, $hAdditionalInfo);
    }

    my $ResponseCode   = $Res->code;
    my $DecodedContent = $Res->decoded_content;

    my $hResLog = {
      Method         => $Method,
      Url            => $Url,
      ResponseCode   => $ResponseCode,
      DecodedContent => $DecodedContent,
      Headers        => $Res->headers_as_string,
      ShopAlias      => $ShopAlias,
      StoreAlias     => $StoreAlias,
    };
    $self->logToCommunication("Response $LogSuffix", $hResLog);

    return $Res;
}

#========================================================================================
# §function     getRDToken
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->getRDToken();
#----------------------------------------------------------------------------------------
# §description  Retrieve an RDT token to access PII information for the SP-API
#----------------------------------------------------------------------------------------
# §return       $RestrictedDataToken | The retrieved restricted data token, or undef in failure | string
#========================================================================================
sub getRDToken {
    my $self = shift;
    my ($Method, $Url, $Content, $hHeaders, $hAdditionalInfo) = @_;

    my $PortalSite = $self->get('PortalSite');
    my $SPAPI_EndPoint = $PortalSite->get('SPAPI_EndPoint');

    my $RDTPath = $hAdditionalInfo->{'RDTPath'};
    my $aRDTElements = exists $hAdditionalInfo->{'RDTElements'} ? $hAdditionalInfo->{'RDTElements'} : [];
    #-- If the body is formatted incorrectly, Amazon may still return you a 'DUD' token, which has no access to anything!!!
    #-- Then we get the dreaded 'Unauthorised' error message which tells us nothing!!
    my $hRestrictedRequest = {
        'restrictedResources' => [{
            'method' => $Method,
            'path'   => $RDTPath,
            'dataElements' => $aRDTElements,
        }],
    };

    my $Response = $self->doRequest(
        'POST',
        $SPAPI_EndPoint . '/tokens/2021-03-01/restrictedDataToken',
        to_json($hRestrictedRequest),
        {'Content-Type' => 'application/json'},
        {'SignRequest' => 1}
    );
    if ( $Response && $Response->is_success() ) {
        my $DecodedContent = $Response->decoded_content;

        my $hJsonResponse;
        eval { $hJsonResponse = decode_json($DecodedContent); };
        $hJsonResponse = undef if $@;

        if ( defined $hJsonResponse ) {
            return $hJsonResponse->{'restrictedDataToken'};
        }
    }

    return undef;
}

#========================================================================================
# §function     getSPAPIAccessToken
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $self->getSPAPIAccessToken();
#----------------------------------------------------------------------------------------
# §description  Retrieve an access token for the SP-API
#----------------------------------------------------------------------------------------
# §return       $AccessToken | The retrieved/generated access token, or undef in failure | string
#========================================================================================
sub getSPAPIAccessToken {
    my $self = shift;

    my $CurrentAccessToken = $self->get('SPAPI_AccessToken');
    my $CurrentAccessTokenExpiry = $self->get('SPAPI_AccessTokenExpiry');

    #-- This is an epoch time, remove a minute because we don't want it to expire just as our code is running ^.^
    $CurrentAccessTokenExpiry = $CurrentAccessTokenExpiry - 60;

    #-- If 'Now' is less than the expiry epoch
    if ( time() <= $CurrentAccessTokenExpiry ) {
        return $CurrentAccessToken;
    }

    my $PortalSite = $self->get('PortalSite');
    my $ClientID = $PortalSite->get('SPAPI_ClientID');
    my $ClientSecret = $PortalSite->get('SPAPI_ClientSecret');
    my $SPAPI_RefreshToken = $self->get('SPAPI_RefreshToken');

    my $Response = $self->doRequest(
        'POST',
        $OAUTH_TOKEN_ENDPOINT,
        "grant_type=refresh_token&refresh_token=$SPAPI_RefreshToken&client_id=$ClientID&client_secret=$ClientSecret",
        {'Content-Type' => 'application/x-www-form-urlencoded;charset=UTF-8'},  #-- Fragile, leave me alone please
        {}
    );
    if ( $Response && $Response->is_success() ) {
        my $DecodedContent = $Response->decoded_content;

        my $hJsonResponse;
        eval { $hJsonResponse = decode_json($DecodedContent); };
        $hJsonResponse = undef if $@;

        #-- We're using the refreshToken to get a new Access Token. It therefore will return the same refresh token, so we only need to set Access Token and Expiry here
        if ( defined $hJsonResponse ) {
            $self->set({
                'SPAPI_AccessToken' => $hJsonResponse->{'access_token'},
                'SPAPI_AccessTokenExpiry' => time() + $hJsonResponse->{'expires_in'},
            });

            return $hJsonResponse->{'access_token'};
        }
    }

    return undef;
}

#========================================================================================
# §function     addProducts
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->addProducts($aProducts);
#----------------------------------------------------------------------------------------
# §description  Creates AmazonOffer items.
#----------------------------------------------------------------------------------------
# §input        $prepareProducts | Products | ref.array.object
#========================================================================================
sub addProducts {
    my $self = shift;
    my ($prepareProducts) = @_;
    $self->SUPER::addProducts($prepareProducts);

    my $Class    = LoadClassByAlias('AmazonOffer');
    my $Currency = $self->get('PortalSite')->get('Currency');
    my $TaxModel = $self->get('PortalSite')->get('TaxModel');
    my $Folder   = $self->folder('AmazonOffers');

    foreach my $Product (@$prepareProducts) {
        my $ParentOffer = undef;
        if ( defined $Product->get('SuperProduct') ) {
            my $ParentOffers =
              $Class->getObjects( 'ByPortalConfigIDProductID',
                [ $self->id, $Product->get('SuperProduct')->id ] );
            $ParentOffer = shift @$ParentOffers;

            # parent offer is missing, so we have to add it first.
            if ( not defined $ParentOffer ) {

                # create ProductPortal assignment.
                $self->SUPER::addProducts( [ $Product->get('SuperProduct') ] );
                $ParentOffer = $self->_addProduct(
                    $Product->get('SuperProduct'),
                    {
                        'Class'                 => $Class,
                        'Folder'                => $Folder,
                        'Currency'              => $Currency,
                        'TaxModel'              => $TaxModel,
                        'IsMissingSuperProduct' => 1
                    }
                );
            }
        }

        $self->_addProduct(
            $Product,
            {
                'Class'       => $Class,
                'Folder'      => $Folder,
                'Currency'    => $Currency,
                'TaxModel'    => $TaxModel,
                'ParentOffer' => $ParentOffer,
            }
        );
    }
}

#========================================================================================
# §function     _addProduct
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->_addProduct($Product, $hInfo);
#----------------------------------------------------------------------------------------
# §description  Adds a product to this Portal and checks if subproducts have to be added
#               as well (recursive). Returns new AmazonOffer object.
# §input        $Product | Product to add | object
# §input        $hInfo   | additional info, used keys are
#               <ul>
#                 <li>Folder      - Folder to add Offer to         - object
#                 <li>Currency    - used currency to get price     - object
#                 <li>TaxModel    - used tax model for this portal - object
#                 <li>Class       - AmazonOffer class (cached)     - object
#               </ul> | ref.hash.*
# §return       $Offer | AmazonOffer | object
#========================================================================================
sub _addProduct {
    my $self    = shift;
    my ($Product, $hInfo) = @_;

    my $Folder                = $hInfo->{'Folder'};
    my $Currency              = $hInfo->{'Currency'};
    my $TaxModel              = $hInfo->{'TaxModel'};
    my $Class                 = $hInfo->{'Class'};
    my $ParentOffer           = $hInfo->{'ParentOffer'};
    my $IsMissingSuperProduct = $hInfo->{'IsMissingSuperProduct'};

    do {
        Warning('DownloadProductsNotSupported',
        {
            'Product'           => $Product->tleHash,
            'ProductAlias'      => $Product->alias,
        });
        return;
    } if ($Product->getWithParent('IsDownloadProduct'));

    # missing super product offers have just been checked for existance.
    if ( !defined $IsMissingSuperProduct ) {
        my $ExistingOffers =
          $Class->getObjects( 'ByPortalConfigIDProductID', [ $self->id, $Product->id ] );

        # capture duplicates.
        do {
            Warning(
                'ProductOfferExists',
                {
                    'Product'           => $Product->tleHash,
                    'ProductAlias'      => $Product->alias,
                    'PortalConfig'      => $self->tleHash,
                    'PortalConfigAlias' => $self->alias,
                    'ShopAlias'         => $self->getSite->alias,
                    'Store'             => GetCurrentDBHandle()->cacheKey,
                }
            );
            return;
        } if ( scalar @$ExistingOffers );
    }

    # add AmazonProduct default values.
    if ( (!defined $Product->getWithParent('AmazonListerID')) or $Product->getSite()->get('IsUnityRegistered') ) {
        $Product->set({'AmazonListerID' => 0});
    }

    # get default values
    my $hDefaultValues = $self->get([
        'DefaultCondition',
        'DefaultConditionNote',
        'DefaultDeliveryLatency',
        'ForceDefaultDeliveryLatency'
    ]);

    $Product->set({'AmazonCondition' => $hDefaultValues->{'DefaultCondition'}});

    if ($hDefaultValues->{'ForceDefaultDeliveryLatency'}) {
        # always use default value in case we have set the related setting
        $Product->set({'AmazonDeliveryLatency' => $hDefaultValues->{'DefaultDeliveryLatency'}});
    } else {
        # get delivery period from product
        my $DeliveryPeriod = $Product->get('DeliveryPeriod');
        if (defined $DeliveryPeriod) {
            # only set the delivery period if it is an integer and fits in the allowed time range
            eval {
                TestInteger('DeliveryPeriod', $DeliveryPeriod, MIN_DELIVERY_LATENCY, MAX_DELIVERY_LATENCY);
            };
            if (ExistsError()) {
                # set delivery latency to undef in case of an error
                $Product->set({'AmazonDeliveryLatency' => undef});
            } else {
                $Product->set({'AmazonDeliveryLatency' => $DeliveryPeriod});
            }
        } else {
            # use default setting in case we have no delivery period set
            $Product->set({'AmazonDeliveryLatency' => $hDefaultValues->{'DefaultDeliveryLatency'}});
        }
    }

    my $Alias = $Product->alias;
    $Alias = $Folder->nextAlias($Alias.'-%d') if ($Folder->existsChild($Alias));

    # add AmazonOffer default values.
    my $OfferInfo = {
        'Alias'          => $Alias,
        'Product'        => $Product,
        'CurrencyID'     => $Currency->{'CurrencyID'},
        'PortalConfig'   => $self,
        'OnlineStatus'   => $AMAZON_ONLINESTATUS{'OFFER_OFFLINE'},
        'IsPriceChanged' => 1,
        'ParentOffer'    => $ParentOffer,
        'AmazonConditionNote' => $hDefaultValues->{'DefaultConditionNote'},
    };


    eval { ## no critic (RequireExistsError)
        my $ProductPrice = $Product->price({
            'CurrencyID' => $Currency->{'CurrencyID'},
            'TaxModel'   => $TaxModel,
        });
        my $hDefaultsPriceAdaption = $self->get([
            'DefaultsPriceAdaption',
            'DefaultsPriceAdaptionIsAbsolute',
            'DefaultsPriceAdaptionValue'
        ]);
        if ($hDefaultsPriceAdaption->{'DefaultsPriceAdaption'} && defined $hDefaultsPriceAdaption->{'DefaultsPriceAdaptionValue'}) {
            if ($hDefaultsPriceAdaption->{'DefaultsPriceAdaptionIsAbsolute'}) {
                $ProductPrice += $hDefaultsPriceAdaption->{'DefaultsPriceAdaptionValue'};
            } else {
                $ProductPrice += $ProductPrice * $hDefaultsPriceAdaption->{'DefaultsPriceAdaptionValue'}/100;
            }
            # round adapted price
            my $FloatFormatter = DE_EPAGES::Core::API::Object::NumberFormat->new;
            my $MinorUnit = GetInfoCurrency($Currency->{'CurrencyID'})->{'MinorUnit'};
            $ProductPrice = $FloatFormatter->round($ProductPrice, $MinorUnit);
        }
        $OfferInfo->{'AmazonPrice'} = $ProductPrice;
    };

    # create offer
    my $Offer = $Folder->insertChild( $Class, $OfferInfo );

    # add children if necessary.
    if ( not defined $IsMissingSuperProduct
        and scalar @{ $Product->folder('SubProducts')->children } )
    {
        $hInfo->{'ParentOffer'} = $Offer;
        foreach my $SubProduct ( @{ $Product->folder('SubProducts')->children } ) {
            $self->_addProduct( $SubProduct, $hInfo );
        }
    }

    return $Offer;
}

#========================================================================================
# §function     hideOffers
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->hideOffers($aOffers);
#----------------------------------------------------------------------------------------
# §description  Hides offers from portal by setting stock to zero.
#----------------------------------------------------------------------------------------
# §input        $aOffers | AmazonOffers | ref.array.object
#========================================================================================
sub hideOffers {
    my $self = shift;
    my ($aOffers) = @_;

    Transaction(
        sub {
            foreach my $Offer (@$aOffers) {
                LogDebug( 'disable Offer:', $Offer->id );
                $Offer->checkNextState( 'disable', {} );
            }
            $self->uploadStock($aOffers);
        }
    );

    return;
}

#========================================================================================
# §function     removeProducts
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $Object->removeProducts( $aProducts );
# §example      $AmazonConfig->removeProducts( [ $Product ] );
#----------------------------------------------------------------------------------------
# §description  Removes a product from Amazon.
#----------------------------------------------------------------------------------------
# §input        $aProducts | Product list | ref.array.object
#========================================================================================
sub removeProducts {
    my $self = shift;
    my ($aProducts) = @_;

    return unless (ref $aProducts eq 'ARRAY');
    my $Class    = LoadClassByAlias('AmazonOffer');

    foreach my $Product (@$aProducts){
        my $ExistingOffers =
          $Class->getObjects( 'ByPortalConfigIDProductID', [ $self->id, $Product->id ] );
        $self->removeOffers($ExistingOffers);
    }

    $self->SUPER::removeProducts($aProducts);
    return;
}

#========================================================================================
# §function     removeOffers
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->removeOffers($aOffers);
#----------------------------------------------------------------------------------------
# §description  Removes Offers from Amazon portal.
#----------------------------------------------------------------------------------------
# §input        $aOffers | AmazonOffers | ref.array.object
#========================================================================================
sub removeOffers {
    my $self = shift;
    my ($aOffers) = @_;
    my $XMLMessageSnippet;
    my $feedQueue;

    my @Products = ();
    foreach my $Offer (@$aOffers) {
        if ( $Offer->get('OnlineStatus') == $AMAZON_ONLINESTATUS{'OFFER_OFFLINE'} ) {
            my $Product = $Offer->get('Product');
            push( @Products, $Product ) if ( defined $Product );
            $Offer->delete;
        } else {
            my $Sku = QuoteXML( $Offer->get('Product')->alias );
            $XMLMessageSnippet = <<END_XML;
  <Product>
    <SKU>$Sku</SKU>
  </Product>
END_XML
            $feedQueue = $self->_queuePostDocument($XMLMessageSnippet, 'Product','_POST_PRODUCT_DATA_','Delete',$Offer);
        }

    }
    $self->SUPER::removeProducts( \@Products );
}

#========================================================================================
# §function     removeAllOffers
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->removeAllOffers;
#----------------------------------------------------------------------------------------
# §description  Removes all Offers from portal.
#========================================================================================
sub removeAllOffers {

    my $self = shift;
    $self->removeOffers( $self->folder('AmazonOffers')->children );

}

#========================================================================================
# §function     processFeedResults
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->processFeedResults();
#----------------------------------------------------------------------------------------
# §description  Runs through the saved feed results and applies changes to the AmazonOffers affected by the feeds
#----------------------------------------------------------------------------------------
# §return       $FeedSummary | Feed processing summary | hash.ref
#               used keys are:
#               <li>FeedsProcessed - number of feeds processed in this run.
#               <li>FeedsPending   - number of feeds still pending.
#========================================================================================
sub processFeedResults {
    my $self        = shift;
    my $FeedSummary = {};
    my $Remote;

    # Due throttling of Amazon API we have to limit the requests per call
    my $MaxFeedResultPerCall = 8;
    my $CountFeedResults = 0;


    my $Class  = LoadClassByAlias('AmazonFeed');
    my $aItems = $Class->getObjects( 'ByPortalConfigIDProcessingStatus',
        [ $self->id, $AMAZON_PROCESSINGSTATUS{'_IN_PROGRESS_'} ] );

    return if ( not scalar @$aItems );
    $Remote = $self->remote();

    # pending offer list is processed at the end.
    my %PendingList = ();

    foreach my $Feed (@$aItems) {

        $CountFeedResults++;
        last if ($CountFeedResults > $MaxFeedResultPerCall);

        ## first get local messages as hash.
        my %LocalMessages = ();
        foreach my $LocalMessage ( @{ $Feed->folder('Messages')->children } ) {
            $LocalMessages{ $LocalMessage->get('AmazonOffer')->get('Product')->alias } =
              $LocalMessage;
        }

        my $StatusRet;
        eval {
            # get Amazon Processing Status.
            $StatusRet = $Remote->getDocumentProcessingStatus( $Feed->get('AmazonReportNumber') );
        };
        # catch '_INVALID_DOCUMENT_TRANSACTION_IDENTIFIER_'.
        if (ExistsError()) {
            my $Error = GetError();
            if ($Error->code eq 'SOAPFault' && $Error->vars()->{'Code'} eq '_INVALID_DOCUMENT_TRANSACTION_IDENTIFIER_' ) {
                # delete old feed locally.
                $Feed->delete;
                next;
            } else {
                Error($Error);
            }
        }

        # Feed is pending... so our Offer is marked pending as well.
        if ( $StatusRet->Result->{'DocumentProcessingStatus'} ne '_DONE_' ) {
            $FeedSummary->{'FeedsPending'}++;
            foreach my $LocalMessage ( values %LocalMessages ) {
                my $Offer = $LocalMessage->get('AmazonOffer');
                LogDebug( 'Offer still has pending feeds:', $Offer->alias );
                $PendingList{$Offer->alias} = $Offer;
            }
            next;
        }

        # Fetch Processing Report.
        my $ProcessingReport;
        eval {
            $ProcessingReport = $Remote->getProcessingReportDocument( $StatusRet->Result->{'DocumentID'} );
        };
        if (ExistsError()) {
            my $Error = GetError();
            # catch '_DOCUMENT_NO_LONGER_AVAILABLE_' and '_ACCESS_TO_DOCUMENT_DENIED_'
            if ($Error->code eq 'SOAPFault' && $Error->vars()->{'Code'} ~~ [qw(_ACCESS_TO_DOCUMENT_DENIED_ _DOCUMENT_NO_LONGER_AVAILABLE_)]) {
                # delete old feed locally.
                $Feed->delete;
                next;
            } else {
                Error($Error);
            }
        }

        my %RemoteErrors  = ();
        my @GeneralErrors = ();
        foreach my $ReportMessage ( @{ $ProcessingReport->Result } ) {
            LogDebug( 'Debug:', $ProcessingReport->Result );

            # find SKU based results.
            foreach my $SKUResult ( WantList( $ReportMessage->{'Result'} ) ) {

                # Report has failed.
                if ( $SKUResult->{'ResultCode'} eq 'Error' ) {
                    if ( not $SKUResult->{'AdditionalInfo'} ) {

                        # failure without sku information.
                        LogDebug( 'general error:', $SKUResult );
                        push @GeneralErrors, $SKUResult;
                        next;
                    }

                    # add to RemoteErrors hash with SKU as key.
                    $RemoteErrors{ $SKUResult->{'AdditionalInfo'}->{'SKU'} } = $SKUResult;

                    # if we find a matching local msesage, save it.
                    my $MatchingMessage = $LocalMessages{ $SKUResult->{'AdditionalInfo'}->{'SKU'} };
                    if ( not defined $MatchingMessage ) {

                        # sku not in local list.
                        LogDebug( 'unmatched sku:', $SKUResult );
                        push @GeneralErrors, $SKUResult;
                        next;
                    }
                    my $ReplacedMessage = $SKUResult->{'ResultDescription'};

                    # format URL as HTML link.
                    $ReplacedMessage =~ s!
                        (http://.*?) # everything starting with http://
                        (\.$|$)      # there could be an extra '.' at the end of the sentence
                        !<a href="$1">$1</a>$2!x;
                    LogDebug( 'sku based failure:', $ReplacedMessage );

                    my $Offer = $MatchingMessage->get('AmazonOffer');

                    # save error to local message.
                    $MatchingMessage->set(
                        {
                            'ErrorCode'    => $SKUResult->{'ResultMessageCode'},
                            # 500 is the maximum length of the error message field.
                            'ErrorMessage' => substr($ReplacedMessage, 0, 500), ## no critic (ProhibitMagicNumbers)
                        }
                    );

                    # set offer as failed.
                    $Offer->checkNextState('processResultFailure');

                    # delete failed element from hash.
                    delete $LocalMessages{ $SKUResult->{'AdditionalInfo'}->{'SKU'} };
                }
            }

            # find offers that are online.
            my $Summary = $ReportMessage->{'ProcessingSummary'};
            # amazon response can contain several summaries!
            # see EPG-15231
            if (ref $Summary eq 'ARRAY') {
                $Summary = pop @{$Summary};
            }
            if ($Summary and $Summary->{'MessagesSuccessful'} ) {
                # we need same number of successful messages as local elements.
                # If number is greater, a local offer has been deleted in the
                # meanwhile. That is ok as well.
                if ( $Summary->{'MessagesSuccessful'} < scalar keys %LocalMessages ) {
                    Warning('OrderReportMessageMismatch',
                    {
                        'ExpectedCount'     => scalar keys %LocalMessages,
                        'ReceivedCount'     => $Summary->{'MessagesSuccessful'},
                        'PortalConfigAlias' => $self->alias,
                        'ShopAlias'         => $self->getSite->alias,
                        'Store'             => GetCurrentDBHandle()->cacheKey,
                    });
                    next;
                }

                # mark all local messages as success
                foreach my $Message ( values %LocalMessages ) {
                    my $Offer = $Message->get('AmazonOffer');
                    # set offer as being online.
                    my $Success = $Offer->checkNextState('processResultSuccess');
                    next unless $Success;

                    # set errors as reviewed.
                    # they can be deleted somewhen in the future, if we
                    # are sure they are not needed anymore.
                    $_->set( { 'IsReviewed' => 1 }) foreach
                        grep { $_->get('IsReviewed') == 0 }
                        @{ LoadClassByAlias('AmazonFeedMessage')->getObjects( 'ByAmazonOfferID', [ $Offer->id ] ) };
                }
            }
        }
        $Feed->set( { 'ProcessingStatus' => $AMAZON_PROCESSINGSTATUS{'_DONE_'} } );
        $FeedSummary->{'FeedsProcessed'}++;
    }

    foreach my $Offer (values %PendingList){
        $Offer->checkNextState('addPending');
    }

    return $FeedSummary;
}

#========================================================================================
# §function     uploadOffers
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->uploadOffers($aOffers)
#----------------------------------------------------------------------------------------
# §description  Generates and uploads the offer feed for a list of AmazonOffers
#----------------------------------------------------------------------------------------
# §input        $aOffers | AmazonOffers | ref.array.object
#========================================================================================
sub uploadOffers {
    my $self      = shift;
    my ($aOffers) = @_;
    my $XMLMessageSnippet = '';
    my $feedQueue;

    foreach my $Offer (@$aOffers) {

        # only matching offers.
        if ( $Offer->get('PortalConfig')->id != $self->id ) {
                Warning(
                    'PortalConfigMismatch',
                    {
                        'AmazonOfferAlias'             => $Offer->alias,
                        'AmazonOfferPortalConfigAlias' => $Offer->get('PortalConfig')->alias,
                        'PortalConfigAlias'            => $self->alias,
                        'ShopAlias'                    => $self->getSite->alias,
                        'Store'                        => GetCurrentDBHandle()->cacheKey,
                    }
                );
            next;
        }

        # only offers that are not online yet.
        next if ( $Offer->get('OnlineStatus') ==
             $AMAZON_ONLINESTATUS{'OFFER_ONLINE'});

        my $Product = $Offer->get('Product');
        my $err = 0;

        # only offers with EAN
        my $UPCEAN = QuoteXML( $Product->getWithParent('UPCEAN') ) // '';
        if ( not defined $UPCEAN ) {
            Warning('OfferUPCEANMissing', {
                'Offer' => $Offer,
            });
            $err++;
        }

        # only offers with ASIN
        my $ASIN = $Offer->get('ASIN');
        if ( not defined $ASIN ) {
            Warning('OfferASINMissing',{
                'Offer' => $Offer,
            });
            $err++;
            next;
        }

        my $hConditionValues = ItemConditionValues( $Product->getWithParent('AmazonCondition') );
        if ( not defined($hConditionValues->{'CompleteString'})) {
            Warning('OfferNoCondition', {
                'Offer'         => $Offer,
                'AmazonOfferID' => $Offer->id,
                'ShopAlias'     => $self->getSite->alias,
                'Store'         => GetCurrentDBHandle()->cacheKey,
            });
            next;
        }

        my $ProductAlias = QuoteXML( $Product->alias );
        my $OfferingCondition = QuoteXML( $hConditionValues->{'CompleteString'} ) ;
        my $ProductId = QuoteXML($ASIN);
        my $Note = QuoteXML( $Offer->get('AmazonConditionNote')//'');

        $XMLMessageSnippet = <<END_XML;

<Product>
    <SKU>$ProductAlias</SKU>

    <StandardProductID>
          <Type>ASIN</Type>
          <Value>$ProductId</Value>
    </StandardProductID>

    <Condition>
        <ConditionType>$OfferingCondition</ConditionType>
        <ConditionNote>$Note</ConditionNote>
    </Condition>

</Product>

END_XML
        # Operation must be "PartialUpdate"
        # "Update" would overwrite all data in amazon, undef fields would be simply deleted.
        $feedQueue = $self->_queuePostDocument( $XMLMessageSnippet, 'Product', '_POST_PRODUCT_DATA_', 'PartialUpdate',$Offer );
    }

    return;
}

#========================================================================================
# §function     uploadStock
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->uploadStock($aOffers)
#----------------------------------------------------------------------------------------
# §description  Generates and uploads the stock feed for a list of AmazonOffers
#----------------------------------------------------------------------------------------
# §input        $aOffers | AmazonOffers | ref.array.object
# §input        $maxStock | maximum limit of stock value | integer
#========================================================================================
sub uploadStock {
    my $self      = shift;
    my ($aOffers, $maxStock) = @_;
    my $XMLMessageSnippet = '';
    my $feedQueue;


    foreach my $Offer (@$aOffers) {
        my $Product      = $Offer->get('Product');
        my $OnlineStatus = $Offer->get('OnlineStatus');
        next if ( $OnlineStatus == $AMAZON_ONLINESTATUS{'OFFER_OFFLINE'} );
        # next if ( $OnlineStatus == $AMAZON_ONLINESTATUS{'OFFER_PROCESSING'} );

        my $Stocklevel =
          $OnlineStatus == $AMAZON_ONLINESTATUS{'OFFER_HIDDEN'}
          ? 0
          : ( $Product->get('StockLevel') // 0 );
        $Stocklevel = (
            defined $maxStock
            ? ( $maxStock < $Stocklevel ? $maxStock : $Stocklevel )
            : $Stocklevel
        );
        # don't transmit stock levels < 0.
        $Stocklevel = 0 if $Stocklevel < 0;

        my $ProductAlias = QuoteXML( $Product->alias );
        my $Latency = $Product->getWithParent('AmazonDeliveryLatency')
                      ? '<FulfillmentLatency>' . $Product->getWithParent('AmazonDeliveryLatency') .
                        '</FulfillmentLatency>'
                      : '';

        $XMLMessageSnippet = <<END_XML;
  <Inventory>
    <SKU>$ProductAlias</SKU>
    <Quantity>$Stocklevel</Quantity>
    $Latency
  </Inventory>
END_XML

        $feedQueue = $self->_queuePostDocument( $XMLMessageSnippet, 'Inventory', '_POST_INVENTORY_AVAILABILITY_DATA_',
                                                'Update', $Offer );
    }
}

#========================================================================================
# §function     uploadPrices
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->uploadPrices($aOffers)
#----------------------------------------------------------------------------------------
# §description  Generates and uploads the price feed for a list of AmazonOffers
#----------------------------------------------------------------------------------------
# §input        $aOffers | AmazonOffers | ref.array.object
#========================================================================================
sub uploadPrices {
    my $self      = shift;
    my ($aOffers) = @_;
    my $XMLMessageSnippet = '';
    my $feedQueue;

    foreach my $Offer (@$aOffers) {

        # skip missing prices.
        next if ( not $Offer->get('AmazonPrice') );
        my $Product = $Offer->get('Product');
        my $ProductAlias = QuoteXML( $Product->alias );
        my $Currency = $Offer->get('CurrencyID') ;
        my $StdPrice = sprintf( '%.2f', $Offer->get('AmazonPrice') );

        $XMLMessageSnippet = <<END_XML;
  <Price>
    <SKU>$ProductAlias</SKU>
    <StandardPrice currency="$Currency">$StdPrice</StandardPrice>
  </Price>
END_XML

        $feedQueue = $self->_queuePostDocument( $XMLMessageSnippet, 'Price', '_POST_PRODUCT_PRICING_DATA_', undef,
                                                $Offer);

    }
}

#========================================================================================
# §function     uploadOrderFulfillment
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->uploadOrderFulfillment($aOrders)
#----------------------------------------------------------------------------------------
# §description  Generates and uploads the order fulfillment feed for a list of CustomerOrders
#----------------------------------------------------------------------------------------
# §input        $aOrders | CustomerOrders | ref.array.object
#========================================================================================
sub uploadOrderFulfillment {
    my $self      = shift;
    my ($aOrders) = @_;
    my $XMLMessageSnippet;

    foreach my $Order (@$aOrders) {
        my $Shipping = $Order->get('LineItemContainer')->get('Shipping');

        # fetch all AmazonLineItems.
        foreach my $LineItem ( grep { $_->instanceOf('LineItemAmazonTransaction') }
            @{ $Order->get('LineItemContainer')->get('LineItems') } )
        {
            my $Transaction = $LineItem->get('AmazonTransaction');
            next unless (defined $Transaction);
            my $DF          = DE_EPAGES::XML::API::DateTimeFormatXML->new;

            my $ExternalOrderId = $Transaction->get('ExternalOrderID');
            my $MerchantFulfillmentID =  $Order->id;
            my $FullfillmentDate = $Order->get('DispatchedOn');
            next unless (defined $FullfillmentDate);
            # subtract 30minutes to avoid timing conflicts: EPG-15376
            $FullfillmentDate =  $FullfillmentDate - DateTime::Duration->new(minutes => 30);
            $FullfillmentDate = $DF->format_datetime( $FullfillmentDate, 'GMT' );
            my $ShippingName = QuoteXML( $Shipping->get('Name') );
            my $Ordercode = QuoteXML( $Transaction->get('ExternalOrderItemCode') );
            my $MerchantFulfillmentItemID = $LineItem->id;
            my $Quantity = sprintf( '%d', $LineItem->get('Quantity') );

            $XMLMessageSnippet = <<END_XML;
  <OrderFulfillment>
    <AmazonOrderID>$ExternalOrderId</AmazonOrderID>
    <MerchantFulfillmentID>$MerchantFulfillmentID</MerchantFulfillmentID>
    <FulfillmentDate>$FullfillmentDate</FulfillmentDate>
    <FulfillmentData>
      <CarrierName>$ShippingName</CarrierName>
      <ShippingMethod>$ShippingName</ShippingMethod>
      <ShipperTrackingNumber></ShipperTrackingNumber>
    </FulfillmentData>
    <Item>
      <AmazonOrderItemCode>$Ordercode</AmazonOrderItemCode>
      <MerchantFulfillmentItemID>$MerchantFulfillmentItemID</MerchantFulfillmentItemID>
      <Quantity>$Quantity</Quantity>
    </Item>
  </OrderFulfillment>
END_XML

           $self->_queuePostDocument( $XMLMessageSnippet, 'OrderFulfillment', '_POST_ORDER_FULFILLMENT_DATA_' );
       }
    }
    $self->_saveOrderFulfillment($aOrders);
}

#========================================================================================
# §function     getOrderDocuments
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->getOrderDocuments($hInput)
#----------------------------------------------------------------------------------------
# §description  Fetches the order documents for a list of OrderReportIDs
#----------------------------------------------------------------------------------------
# §input        $hInput | parameters for this function (like: Reports, StartDate, EndDate) | ref.hash
#========================================================================================
sub getOrderDocuments {
    my $self           = shift;
    my ($hInput) = @_;

    my $aOrderReportIDs = $hInput->{'Reports'};

    my $Remote = $self->remote();

    # get transactions
    my $AmazonTransaction = DE_EPAGES::Amazon::API::AmazonTransaction->new;

    my $Handler;
    my @CompletedReports = ();
    my $ahReports = [];
    my $hParams = {};

    my $StartDate = undef;
    my $EndDate = undef;

    # do we have a StartDate parameter
    if($hInput->{'StartDate'}){
        # parse for verifying and set UTC timezone
        $StartDate = DateTime::Format::ISO8601->new()->parse_datetime($hInput->{'StartDate'});
        $StartDate->set_time_zone('UTC');
    } else {
        $StartDate = $self->get("LastReportExportDate") ? $self->get("LastReportExportDate") : undef;
        if (defined $StartDate) {
            $StartDate->set_time_zone('UTC');
        }
    }

    # do we have a StartDate parameter
    if($hInput->{'EndDate'}){
        # parse for verifying and set UTC timezone
        $EndDate = DateTime::Format::ISO8601->new()->parse_datetime($hInput->{'EndDate'});
        $EndDate->set_time_zone('UTC');
    } else {
        $EndDate = defined $StartDate ? AddDuration($StartDate, (keys %REPORTS_TIME_SPAN)[0], $REPORTS_TIME_SPAN{(keys %REPORTS_TIME_SPAN)[0]}) : undef;
        if (defined $EndDate) {
            $EndDate->set_time_zone('UTC');
        }
    }

    if (defined $StartDate) {
        $hParams->{'createdSince'} = $StartDate->datetime();
    }
    if (defined $EndDate) {
        $hParams->{'createdUntil'} = $EndDate->datetime();
    }

    # do we have a EndDate parameter
    if($hInput->{'EndDate'}) {
        # parse for verifying and set UTC timezone
        $EndDate = DateTime::Format::ISO8601->new()->parse_datetime($hInput->{'EndDate'});
        $EndDate->set_time_zone('UTC');
        $hParams->{'createdUntil'} = $EndDate->datetime();
    }

    # get reports for given types and parameters
    $hParams->{'reportTypes'} = $AMAZON_PENDING_FEEDS{'orders'};
    $ahReports = $Remote->getAllPendingDocumentInfo(
        $hParams,
        $aOrderReportIDs,
    );

    my $Today = DateTime->now();
    my $DeltaDaysFromToday = defined $StartDate ? $StartDate->delta_days($Today)->delta_days : undef;
    my $NextStartDate = defined $StartDate ? AddDuration($StartDate, 'days', 1) : undef;
    # if reports empty and the startDate is not the same as today
    # we set LastReportExportDate to +1 Day
    # e.g: Scan: startDate 01-01-2022 endDate 01-03-2022 -> nothing found LastReportExportDate=01-02-2022
    if ( (not scalar(@{$ahReports->{'reports'}})) && ( (defined $DeltaDaysFromToday) && ($DeltaDaysFromToday != 0) ) ) {
        $self->set({"LastReportExportDate" => $NextStartDate->datetime() });
        return;
    }

    my $ExistsReports = 0;
    # we need to sort the reports beginning with the oldest report
    my $ahsortedReports = [sort {DateTime::Format::ISO8601->new()->parse_datetime($a->{'createdTime'}) <=> DateTime::Format::ISO8601->new()->parse_datetime($b->{'createdTime'})}  @{$ahReports->{'reports'}}];

    foreach my $hReport ( @{$ahsortedReports} ) {
        my $ReportDocId = $hReport->{'reportDocumentId'};
        my $ReportId = $hReport->{'reportId'};
        my $ReportCreatedTime = DateTime::Format::ISO8601->new()->parse_datetime($hReport->{'createdTime'});
        $ReportCreatedTime->set_time_zone('UTC');

        # the startDate is not the same as today
        # and if the report we want to import has same date as startDate we have this one already
        # if we have newer ones this will be override with the next item in the loop
        if ( ((defined  $DeltaDaysFromToday) && ($DeltaDaysFromToday != 0)) && ($ReportCreatedTime->datetime() eq $StartDate->datetime()) ) {
            $self->set({ "LastReportExportDate" => $NextStartDate->datetime() });
            next;
        }

        eval { $Handler = $Remote->getOrderDocument( $ReportDocId, $ReportId ); };

        if ( ExistsError() ) {
            my $Error = GetError();
            $Error->addVars({ 'Path' => $self->pathString });
            Warning( $Error );
        } else {
            my $aTransactions = $Handler->Result;
            my @CompletedTransactions = $AmazonTransaction->insert( $self, $aTransactions );
            if ( scalar @CompletedTransactions == scalar @$aTransactions ) {
                push @CompletedReports, $ReportDocId;
                if(!$ExistsReports){$ExistsReports = 1};
                # make sure you set this date only if the export was successfully finished
                # we need to convert it back to Datetime format to save it
                # don't set date when triggering single reports or a date range was given
                unless ((defined $aOrderReportIDs) || $hInput->{'StartDate'} || $hInput->{'EndDateDate'}) {
                    $self->set({"LastReportExportDate" => $ReportCreatedTime->datetime() });
                }
            }
            else {
                Warning(
                    'OrderCountMismatch',
                    {
                        'ImportOrderCount'    => scalar @CompletedTransactions,
                        'XMLTransactionCount' => scalar @$aTransactions,
                        'ReportNumber'        => $ReportId,
                        'PortalConfigAlias'   => $self->alias,
                        'Path'                => $self->pathString,
                        'ShopAlias'           => $self->getSite->alias,
                        'Store'               => GetCurrentDBHandle()->cacheKey,
                    }
                );
            }
        }

    }

    return if ( !$ExistsReports );

    # upload shop order number to amazon (only prepared here and triggered by submitFeedQueue.pl )
    $self->uploadOrderAcknowledgement( $AmazonTransaction->getImportedLineItems() );
}

#========================================================================================
# §function     uploadOrderAcknowledgement
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->uploadOrderAcknowledgement($aOrderLineItems)
#----------------------------------------------------------------------------------------
# §description  Generates and uploads a feed for acknowledging the receipt of a list of offers
#----------------------------------------------------------------------------------------
# §input        $aOrderLineItems | LineItemAmazonTransactions | ref.array.object
#========================================================================================
sub uploadOrderAcknowledgement {
    my $self = shift;

    my ($aOrderLineItems) = @_;
    my $XMLMessageSnippet;

    return if ( not defined $aOrderLineItems );

    foreach my $OrderLineItem (@$aOrderLineItems) {
        my $Order             = $OrderLineItem->parent->parent;
        my $AmazonTransaction = $OrderLineItem->get('AmazonTransaction');
        next if ( !defined $AmazonTransaction );

        my $AmazonOrderID = $AmazonTransaction->get('ExternalOrderID');
        my $MerchantOrderID = QuoteXML( $Order->alias );
        my $AmazonOrderItemCode = $AmazonTransaction->get('ExternalOrderItemCode');
        my $MerchantOrderItemID = $OrderLineItem->id;

        $XMLMessageSnippet = <<END_XML;
    <OrderAcknowledgement>
        <AmazonOrderID>$AmazonOrderID</AmazonOrderID>
        <MerchantOrderID>$MerchantOrderID</MerchantOrderID>
        <StatusCode>Success</StatusCode>
        <Item>
            <AmazonOrderItemCode>$AmazonOrderItemCode</AmazonOrderItemCode>
            <MerchantOrderItemID>$MerchantOrderItemID</MerchantOrderItemID>
        </Item>
    </OrderAcknowledgement>
END_XML

        $self->_queuePostDocument($XMLMessageSnippet, 'OrderAcknowledgement', '_POST_ORDER_ACKNOWLEDGEMENT_DATA_');
    }
}

#========================================================================================
# §function     postDocument
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->_postDocument($XML,$MessageType,$FeedType);
#----------------------------------------------------------------------------------------
# §description  generates the AmazonEnvelope XML Document and posts it to Amazon
#----------------------------------------------------------------------------------------
# §input        $XMLMessages | xml content             | string
# §input        $MessageType | amazon message type     | string
# §input        $FeedType    | feed type name          | string
# §return       $Response    | SOAP response           | object
#========================================================================================
sub postDocument {
    my $self        = shift;
    my ($XMLMessages, $MessageType, $FeedType) = @_;

    my $Remote = $self->remote();

    my $MerchantIdentifier = $self->get('MWSMerchantID') // '';
    my $XML = <<END_XML;
<?xml version="1.0"?>
<AmazonEnvelope xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
  xsi:noNamespaceSchemaLocation="amzn-envelope.xsd">
  <Header>
    <DocumentVersion>1.01</DocumentVersion>
      <MerchantIdentifier>$MerchantIdentifier</MerchantIdentifier>
    </Header>
  <MessageType>$MessageType</MessageType>
  $XMLMessages
</AmazonEnvelope>
END_XML

    return $Remote->postDocument( $FeedType, $XML );
}

#========================================================================================
# §function     _queuePostDocument
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->_queuePostDocument($XMLMessageSnippet,$MessageType,
#                                                   $FeedType,$OperationType,$AmazonOffer);
#----------------------------------------------------------------------------------------
# §description  Buffers the post to Amazon to the submitFeedQueue. A scheduler runs the
#               the script submitFeedQueue.pl which calls the postDocument function with
#               a bundled call
#----------------------------------------------------------------------------------------
# §input        $XMLMessages    | xml content                       | string
# §input        $MessageType    | amazon message type               | string
# §input        $FeedType       | feed type name                    | string
# §input        $OperationType  | Operation type name               | string
# §input        $AmazonOffer    | Related Amazon offer (optional)   | object
# §return       $feedQueue      | submitFeedQueue                   | object
#========================================================================================
sub _queuePostDocument{
    my $self        = shift;
    my ($XMLMessageSnippet, $MessageType, $FeedType, $OperationType, $AmazonOffer) = @_;

    my $Class=LoadClassByAlias('AmazonSubmitFeedQueue');

    if(defined $AmazonOffer) {
        $Class->deleteObjects( 'ByPortalConfigIDFeedTypeAmazonOfferID',[ $self->id, $FeedType, $AmazonOffer->id ] );
    }

    my $feedQueue = $Class->insertObject ({
       'PortalConfig' => $self,
       'XMLMessageSnippet'    => $XMLMessageSnippet,
       'MessageType' => $MessageType,
       'FeedType'        => $FeedType,
       'OperationType'   => $OperationType,
       'AmazonOffer' => $AmazonOffer,
    });

    if( defined $AmazonOffer
        and not($MessageType eq 'Inventory')
        and not($AmazonOffer->get('OnlineStatus') == $AMAZON_ONLINESTATUS{'OFFER_HIDDEN'})
       ){
        $AmazonOffer->set({'OnlineStatus' => ($AMAZON_ONLINESTATUS{'OFFER_PROCESSING'})});
    }

    return $feedQueue;
}

#========================================================================================
# §function     _saveUploadMessages
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->_saveUploadMessages($MessageType,$FeedType,$DocumentID,$Offers);
#----------------------------------------------------------------------------------------
# §description  Saves Amazon processing results.
#----------------------------------------------------------------------------------------
# §input        $MessageType | message type            | string
# §input        $FeedTypeID  | feed type id            | integer
# §input        $DocumentID  | amazon document id      | string
# §input        $aOffers     | list of affected offers | ref.array
#========================================================================================
sub saveUploadMessages {
    my $self        = shift;
    my ($MessageType, $FeedTypeID, $DocumentID, $aOffers) = @_;

    my $FeedClass    = LoadClassByAlias('AmazonFeed');
    my $MessageClass = LoadClassByAlias('AmazonFeedMessage');

    my $Folder = $self->folder('AmazonFeeds');

    my $hInfoFeed = {
        'Alias'              => 'AmazonFeed_' . $MessageType . '_' . $DocumentID,
        'AmazonReportNumber' => $DocumentID,
        'FeedTypeID'         => $FeedTypeID,
        'PortalConfig'       => $self,
        'ListingTimestamp'   => GetCurrentDateTime(),
        'ProcessingStatus'   => $AMAZON_PROCESSINGSTATUS{'_IN_PROGRESS_'}
    };

    my $Feed = $Folder->insertChild( $FeedClass, $hInfoFeed );

    my $MessageFolder = $Feed->folder('Messages');

    foreach my $Offer (@$aOffers) {
        my $hInfoMsg = {
            'Alias'       => 'AmazonFeedMessage_' . $DocumentID . '_' . $Offer->alias,
            'AmazonFeed'  => $Feed,
            'AmazonOffer' => $Offer,
            'IsReviewed'  => 0,
        };
        $Offer->set( { 'OnlineStatus' => $AMAZON_ONLINESTATUS{'OFFER_PROCESSING'}, } );
        $MessageFolder->insertChild( $MessageClass, $hInfoMsg );
    }
}

#========================================================================================
# §function     _saveOrderFulfillment
# §state        private
#----------------------------------------------------------------------------------------
# §syntax       $AmazonConfig->_saveOrderFulfillment($Offers);
#----------------------------------------------------------------------------------------
# §description  Saves DispatchedOn to Amazon Fulfillment Timestamp
#----------------------------------------------------------------------------------------
# §input        $aOrders | CustomerOrders | ref.array.object
#========================================================================================
sub _saveOrderFulfillment {
    my $self = shift;
    my ($aOrders) = @_;

    foreach my $Order (@$aOrders) {
        foreach my $LineItem ( grep { $_->instanceOf('LineItemAmazonTransaction') }
            @{ $Order->get('LineItemContainer')->get('LineItems') } )
        {
            my $Transaction = $LineItem->get('AmazonTransaction');
            next unless (defined $Transaction);
            $Transaction->set( { 'FulfillmentTimestamp' => $Order->get('DispatchedOn') } )
              if ( defined $Transaction );
        }
    }
}

#========================================================================================
# §function     updateBestPrice
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->updateBestPrice( $aOffers );
#----------------------------------------------------------------------------------------
# §description  Update the Best Price at Amazon for the given offers
#----------------------------------------------------------------------------------------
# §input        $aOffers | list of offers to update | ref.array
#========================================================================================
sub updateBestPrice {
    my $self = shift;
    my ($aOffers) = @_;

    my $PortalSite = $self->get('PortalSite');
    my $Portal = $self->get('Portal');

    my %ASINs = ();

    # Get the update interval which is set in the PBO
    my $UpdateInterval = $Portal->get('AWSPriceUpdateInterval');

    my $CurrentDateTime = GetCurrentDBHandle()->currentDateTime;

    foreach my $Offer (@$aOffers) {
        my $LastUpdate = $Offer->get('LastBestPriceUpdate');

        # Update only outdated offers
        next if ( defined $LastUpdate and ( $CurrentDateTime->epoch() - $LastUpdate->epoch() ) < $UpdateInterval );
        $ASINs{ $Offer->get('ASIN') } = $Offer if $Offer->get('ASIN');
        # stop if maximum request quota is reached
        last if ( ( scalar keys %ASINs ) == MAX_PRICEUPDATES_PER_CALL );
    }

    # No updates to be done
    return if ( !scalar keys %ASINs );

    my $Remote = $self->remote();
    my @ASINList = keys %ASINs;

    # Get the prices for the requested ASINs
    my $aLowestPriceList = $Remote->getLowestOfferList($PortalSite->get('MWSMarketplaceID'), \@ASINList )->{'Result'};

    # Save all updated prices to the offer wich are associated with the ASINs
    foreach my $Item ( @$aLowestPriceList ){
        my $CurrentItem = $ASINs{$Item->{'ASIN'}};
        my $hValues = {
            'BestPrice' => $Item->{'LowestPrice'},
            'LastBestPriceUpdate' => $CurrentDateTime,
        };
        $CurrentItem->set($hValues);
    }
}


#========================================================================================
# §function     remote
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->remote( );
#----------------------------------------------------------------------------------------
# §description  Instanciates a RemoteAdapter using RemoteAdapterFactory
#----------------------------------------------------------------------------------------
# §return       $Remote | remote adapter | object
#========================================================================================
sub remote {
    my $self = shift;
    my $Remote;
    eval {
        if ($self->get('UseMWSAccount')) {
            $Remote = DE_EPAGES::Amazon::API::RemoteAdapterFactory::NewRemoteMerchantToMWSAdapter($self);
        }
    };
    if (ExistsError()) {
        my $Error = GetError();
        Error($Error) unless ($Error->code eq 'MissingConstructorArgument');
        Error('AmazonMerchantSettingsIncomplete', {
            'ErrorMessage' => $Error->message(),
        }, $Error);
    }

    return $Remote;
}

#========================================================================================
# §function     activateOrderScheduler
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->activateOrderScheduler();
#----------------------------------------------------------------------------------------
# §description  sets order scheduler to every 15 minutes, only if supported by remote adapter.
#========================================================================================
sub activateOrderScheduler {
    my $self = shift;
    $self->remote->activateOrderScheduler(REPORT_SCHEDULER_PERIOD);
}

#========================================================================================
# §function     getOrderSchedulers
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->getOrderSchedulers();
#----------------------------------------------------------------------------------------
# §description  get order scheduler ids.
#========================================================================================
sub getOrderSchedulers {
    my $self = shift;
    my $aSchedulerIDs = $self->remote->getOrderSchedulers();

    return $aSchedulerIDs;
}

#========================================================================================
# §function     deleteOrderScheduler
# §state        public
#----------------------------------------------------------------------------------------
# §syntax       $self->deleteOrderScheduler($aSchedulerIDs);
#----------------------------------------------------------------------------------------
# §description  delete order scheduler.
#========================================================================================
sub deleteOrderScheduler {
    my $self = shift;
    my ($SchedulerID) = @_;

    $self->remote->deleteOrderScheduler($SchedulerID);
}

1;
