use strict;
use Getopt::Long;
use Time::HiRes qw (time);
use DE_EPAGES::Amazon::API::Constants qw ( SP_API_REPORTS_RESSOURCE %AMAZON_PENDING_FEEDS );
use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::Error qw ( ExistsError GetError );
use DE_EPAGES::Core::API::Script qw ( RunScript );
use DE_EPAGES::Core::API::Warning qw ( Warning );
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Core::API::Url            qw ( BuildParameterString );
use DE_EPAGES::Database::API::Connection qw ( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath LoadClassByAlias );
use DE_EPAGES::Shop::API::Feature qw( LoadFeatureByAlias );
use DE_EPAGES::WebInterface::API::MessageCenter qw ( SynchronizeCache );
use DE_EPAGES::Amazon::API::RemoteAdapterFactory;
use JSON;

sub Main {

    local $| = 1;

    my $Password;
    my $StoreName;
    my $AmazonPlatformAlias = '';
    my $ShopAlias;
    my $Help;
    my $EndDate;
    my $StartDate;

    GetOptions(
        'help'        => \$Help,
        'passwd=s'    => \$Password,
        'storename=s' => \$StoreName,
        'platform=s'  => \$AmazonPlatformAlias,
        'shop=s'      => \$ShopAlias,
        'startdate=s' => \$StartDate,
        'enddate=s'   => \$EndDate,

    );

    usage() if $Help;
    usage() unless $StoreName;

    my $time1 = time;

    RunOnStore(
        Store      => $StoreName,
        DBPassword => $Password,
        Sub        => sub {
            GetReportsCount( {
                    'AmazonPlatformAlias' => $AmazonPlatformAlias,
                    'ShopAlias'           => $ShopAlias,
                    'StartDate'           => $StartDate,
                    'EndDate'             => $EndDate,
                    'StoreName'           => $StoreName,
                }
            );
        },
    );
    my $time2   = time;
    my $seconds = $time2 - $time1;
    my $logmsg  = sprintf( 'need %.3f seconds', $seconds );
    print $logmsg. "\n";
    GetLog()->debug("Amazon Order Documents: $logmsg");

    eval { SynchronizeCache(); };
    Warning( GetError() ) if ExistsError();
}

sub GetReportsCount {    ## no critic (ProhibitFunctionSize)
    my ($hInput) = @_;

    # make sure that the cartridge is installed
    unless ( ExistsCartridgeByPackage('DE_EPAGES::Amazon') ) {
        print 'Store: ' . $hInput->{'StoreName'} . ": Cartridge Amazon is not installed.\n";
        return;
    }

    my @AmazonConfigs     = ();
    my $AmazonConfigClass = LoadClassByAlias('AmazonConfig');
    if ( defined $hInput->{'ShopAlias'} and $hInput->{'ShopAlias'} ne 'all' ) {
        my $Shop;
        eval { $Shop = LoadObjectByPath('/Shops')->child( $hInput->{'ShopAlias'} ); };
        if ( ExistsError() ) {
            printf STDERR ( "[%s] WARN: Shop '%s' does not exist.\n", $hInput->{'StoreName'}, $hInput->{'ShopAlias'} );
            return;
        }

        my $aPortalConfigs = $Shop->get('PortalConfigs');
        @AmazonConfigs = grep { $_->instanceOf($AmazonConfigClass) } @{$aPortalConfigs};
    }
    else {
        @AmazonConfigs = @{ $AmazonConfigClass->objects };
    }

    my ($StartDate, $EndDate);
    if($hInput->{'StartDate'}){
        # parse for verifying and set UTC timezone
        $StartDate = DateTime::Format::ISO8601->new()->parse_datetime($hInput->{'StartDate'});
        $StartDate->set_time_zone('UTC');
        $StartDate = $StartDate->datetime();
    }
    if($hInput->{'EndDate'}){
        # parse for verifying and set UTC timezone
        $EndDate = DateTime::Format::ISO8601->new()->parse_datetime($hInput->{'EndDate'});
        $EndDate->set_time_zone('UTC');
        $EndDate = $EndDate->datetime();
    }

    # check if feature is enabled.
    my $Feature = LoadFeatureByAlias('Amazon');
    foreach my $AConfig (@AmazonConfigs) {
        my $PortalSite   = $AConfig->get('PortalSite');
        my $Shop         = $AConfig->getSite;
        my $FeatureValue = $Shop->featureMaxValue($Feature);

        next if ( not $FeatureValue );
        # skip if user requested a particular site.
        next if ( $hInput->{'AmazonPlatformAlias'} && $hInput->{'AmazonPlatformAlias'} ne $PortalSite->alias );
        # skip if shop is ClosedByProvider
        next if ($Shop->get('ClosedByProvider'));
        # skip if no SPAPI_AccessToken is set
        next unless defined $AConfig->get('SPAPI_AccessToken');


        my $SP_API_ENDPOINT = $PortalSite->get('SPAPI_EndPoint');
        my $URI = $SP_API_ENDPOINT . SP_API_REPORTS_RESSOURCE . '/reports' ;

        my $hContent = {
            'reportTypes'        => $AMAZON_PENDING_FEEDS{'orders'},
            'pageSize'           => 100,
            'processingStatuses' => 'DONE',
        };

        $hContent->{'createdSince'} = $StartDate if defined $StartDate;
        $hContent->{'createdUntil'} = $EndDate if defined $EndDate;

        $URI = _buildRequestURL($URI, $hContent);
        my $Response = $AConfig->doRequest('GET', $URI, '', {'Content-Type' => 'application/json'}, {'SignRequest'=>1});
        my $hReports = {};
        $hReports = from_json($Response->decoded_content);

        if((ref $hReports eq 'HASH') and exists $hReports->{'reports'} ){
            print 'Found '.scalar(@{$hReports->{'reports'}}).' reports in one API call for Store:'.$hInput->{'StoreName'}.
                " ShopID:".$Shop->id."(".$Shop->alias.") AmazonConfig:".$AConfig->id.".\n";
            if ($hReports->{'nextToken'}){
                print 'Paging detected for Store:'.$hInput->{'StoreName'}.
                    " ShopID:".$Shop->id."(".$Shop->alias.") AmazonConfig:".$AConfig->id.".\n\n"
            }
        } else{
            print '(unclear reports response) No reports found for Store:'.$hInput->{'StoreName'}.
                " ShopID:".$Shop->id."(".$Shop->alias.") AmazonConfig:".$AConfig->id.".\n";
        }
     }
}

sub _buildRequestURL{
    my ($APIUrl, $hQueryParams) = @_;

    my $ParamString = defined $hQueryParams ? BuildParameterString($hQueryParams) . '&' : '';
    $APIUrl .= "?$ParamString";

    return $APIUrl;
}



sub usage {
    print <<END_USAGE;
Description:
    This script prints out all shops which have more than 100 reports which results in pagination.

Usage:
    perl $0 [options] [flags]

options:
    -passwd       database user password              ()
    -storename    name of store                       (Store)
    -shop         shop alias             (optional)   (all)
    -platform     amazon platform        (optional)   (Amazon_de_DE)
    -startdate    start date (ISO 8601) for fetching order documents (optional but mandatory if -enddate is given)
    -enddate      end date (ISO 8601) for fetching order documents (optional but mandatory if -startdate is given)
                  ! use UTC time for startdate and enddate as ISO 8601 value. Have a look to http://www.timestamp-converter.com/
                  for converting your TZ to UTC
flags:
    -help       show the command line options

Example:
    perl $0 -storename Store
    perl $0 -storename Store -startdate 2022-07-27T07:00:00.000Z -enddate 2022-07-27T08:00:00.000Z
    perl $0 -storename Store -shop DemoShop -platform Amazon_de_DE -startdate 2022-07-27T07:00:00.000Z -enddate 2022-07-27T08:00:00.000Z
END_USAGE
    exit 2;
}

RunScript( 'Sub' => \&Main );
