#========================================================================================
# package      DE_EPAGES::Core::API::REST::BaseExportDriver
# state        public
#----------------------------------------------------------------------------------------
# description  Base class for REST export drivers
#========================================================================================
package DE_EPAGES::Core::API::REST::BaseExportDriver;

use strict;

use DE_EPAGES::Core::API::Iterator qw( ArrayIterator );


#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Exporter = DE_EPAGES::Core::API::REST::BaseExportDriver->new(%options);
#----------------------------------------------------------------------------------------
# description  create a new REST export driver
#----------------------------------------------------------------------------------------
# input        %options |  exporter options | hash
# return       $Exporter |  exporter | object
#========================================================================================
sub new {
    my $class = shift;
    my %params = @_;
    return bless \%params, $class;
}

#========================================================================================
# function     exportREST
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Exporter->exportREST($FileName, $hFormat, $aColumns, $Items );
#----------------------------------------------------------------------------------------
# description  exports data (2 dim. array) into a file.
#               starts the document, calls function parse and closes the document
#----------------------------------------------------------------------------------------
# input        $FileName |  export file | string
# input        $hFormat | export file format
#                   <ul>
#                     <li>eol - end of line (optional, default: "\n")</li>
#                     <li>encoding - character encoding (optional, default: "iso-8859-1")</li>
#                     <li>quote_char - field delimiter (optional, default: '"')</li>
#                     <li>escape_char - escape character (optional, default: '"')</li>
#                     <li>sep_char - field separator (optional, default: ",")</li>
#                     <li>open_mode - file open mode (optional, default: ">")</li>
#                     <li>usebom - write a byte order mark for unicode files if 'utf8' or
#                         encoding  =~ /^utf/i  (optional, default: 1)</li>
#                   </ul>
#                    | ref.hash
# input        $aColumns | column names (used for header) | ref.array
# input        $Items | data for export | DE_EPAGES::Core::API::Iterator or ref.array.array.string
#========================================================================================
sub exportREST {
    my $self = shift;
    my ($aColumns, $Items) = @_;

    $self->{'LineNumber'} = 0;
    $self->{'Columns'}    = $aColumns;
    $self->{'Items'}      = $Items // [];
    $self->{'NoHeader'}   = 1;

    $self->drive();

    return;
}

#========================================================================================
# function     items
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Items = $self->items();
#----------------------------------------------------------------------------------------
# description  return the items to export
#----------------------------------------------------------------------------------------
# return       $Items | parameter of exportREST | ref.array.array.string or DE_EPAGES::Core::API::Iterator
#========================================================================================
sub items { return shift->{'Items'}; }

#========================================================================================
# function     iterator
# state        public
#----------------------------------------------------------------------------------------
# syntax       my $Items = $self->iterator;
# example      my $Items = $Driver->iterator;
#               while ( my $Item = <$Items> ) {
#                   ProcessItem( $Item );
#               }
#----------------------------------------------------------------------------------------
# description  returns an iterator for items to export
#----------------------------------------------------------------------------------------
# return       $Items | item iterator | ref.array.array.string or DE_EPAGES::Core::API::Iterator
#========================================================================================
sub iterator {
    my $self = shift;

    my $Items = $self->items;

    return (ref($Items) eq 'ARRAY')
        ? ArrayIterator( [ @$Items ] ) # copy the array, because the iterator would destroy it
        : $Items;
}

#========================================================================================
# function     columns
# state        public
#----------------------------------------------------------------------------------------
# syntax       $aColumns = $self->columns;
#----------------------------------------------------------------------------------------
# description  return the column name array
#----------------------------------------------------------------------------------------
# return       $aColumns
#========================================================================================
sub columns { return shift->{'Columns'}; }

#========================================================================================
# function     incLineNumber
# state        public
#----------------------------------------------------------------------------------------
# syntax       $linenumber = $self->incLineNumber;
#----------------------------------------------------------------------------------------
# description  return the incremented linenumber
#----------------------------------------------------------------------------------------
# return       $linenumber |
#========================================================================================
sub incLineNumber { return ++shift->{'LineNumber'}; }

#========================================================================================
# function     lineNumber
# state        public
#----------------------------------------------------------------------------------------
# syntax       $linenumber = $self->lineNumber;
#----------------------------------------------------------------------------------------
# description  return the current linenumber
#----------------------------------------------------------------------------------------
# return       $linenumber |
#========================================================================================
sub lineNumber { return shift->{'LineNumber'}; }

#========================================================================================
# function     writeLine
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->writeLine($aLineElements);
#----------------------------------------------------------------------------------------
# description  exports a line ( converts and writes to the filehandle )
#----------------------------------------------------------------------------------------
# input        $aLineElements | list of strings for 1 line | ref.array.string
#========================================================================================
sub writeLine {
    my $self = shift;
    my ($aLineElements)  = @_;

    $self->_print($aLineElements);
    $self->incLineNumber;
    return;
}

sub _print {
    my $self = shift;
    my ($aLineElements) = @_;

    $self->exportItem($aLineElements);
    return;
}

#========================================================================================
# function     exportLine
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->exportLine();
#----------------------------------------------------------------------------------------
# description  exports one data line, get the line element from function topID('Item')
#========================================================================================
sub exportLine {
    my $self = shift;

    my $aValues = $self->topID('Item');
    $self->writeLine( $aValues );
    return;
}

#========================================================================================
# function     writeLines
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->writeLines();
#----------------------------------------------------------------------------------------
# description  writes the content lines
#               each item of $self-><function iterator> is exported via <function exportLine>.
#========================================================================================
sub writeLines {
    my $self = shift;

    $self->onStartExport();

    my $Items = $self->iterator;
    while( my $Item = <$Items> ) {
        $self->pushID('Item', $Item);
        $self->exportLine();
        $self->popID('Item');
    }

    $self->onEndExport();
    return;
}

#========================================================================================
# function     drive
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->drive();
#----------------------------------------------------------------------------------------
# description  write the data content lines
#========================================================================================
sub drive {
    my $self = shift;

    $self->writeLines();
    return;
}

#========================================================================================
# function     pushID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->pushID( $Key, $Value );
# example      $self->pushID( 'CompanyID', $CompanyID );
#----------------------------------------------------------------------------------------
# description  adds a parameter to the parameter stack. Any previous parameter with the
#               same name will be hidden until $self->popID( $Key ) is called.
#               Use $self->topID( $Key ) to get the parameter value.
#               Use $self->popID( $Key ) to remove the parameter from the stack.
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# input        $Value | parameter value
#========================================================================================
sub pushID {
    my $self = shift;
    my ($Key, $Value) = @_;

    $self->{'IDs'}->{$Key} = [] unless exists $self->{'IDs'}->{$Key};
    push @{$self->{'IDs'}->{$Key}}, $Value;
}

#========================================================================================
# function     popID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $self->popID( $Key );
# example      $CompanyID = $self->popID( 'CompanyID' );
#----------------------------------------------------------------------------------------
# description  removes a parameter from the parameter stack and returns its value.
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# return       $Value | parameter value
#========================================================================================
sub popID {
    my $self = shift;
    my ($Key) = @_;

    return pop @{$self->{'IDs'}->{$Key}};
}

#========================================================================================
# function     topID
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Value = $self->topID( $Key );
# example      $CompanyID = $self->topID( 'CompanyID' );
#----------------------------------------------------------------------------------------
# description  returns the current value of a parameter
#----------------------------------------------------------------------------------------
# input        $Key | parameter name | string
# return       $Value | parameter value
#========================================================================================
sub topID {
    my $self = shift;
    my ($Key) = @_;

    return () unless exists $self->{'IDs'}->{$Key};
    return $self->{'IDs'}->{$Key}->[-1];
}

#========================================================================================
# function     onStartExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->onStartExport();
#----------------------------------------------------------------------------------------
# description  onStartExport gets called before the $Items iteration starts
#========================================================================================
sub onStartExport {
    my $self = shift;
    die "method onStartExport not implemented in $self";
}

#========================================================================================
# function     onEndExport
# state        public
#----------------------------------------------------------------------------------------
# syntax       $self->onEndExport();
#----------------------------------------------------------------------------------------
# description  onEndExport gets called after the $Items iteration ends
#========================================================================================
sub onEndExport {
    my $self = shift;
    die "method onEndExport not implemented in $self";
}

1;
