#========================================================================================
# package      DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentType
# state        public
# base         DE_EPAGES::Shop::API::Object::ShopObject
# description  object interface for ShippingFulfillmentType.
#========================================================================================
package DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentType;
use base DE_EPAGES::Shop::API::Object::ShopObject;

use strict;
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Core::API::PerlTools qw (isValid GetLocalTimeZone);
use DE_EPAGES::Trigger::API::Trigger qw ( TriggerHook );
use DE_EPAGES::ExternalShipping::API::Shop qw (GetShippingFulfillmentMethodsFolder GetShippingProviderMethodsFolder);
use DateTime;

#========================================================================================
# function     className
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ClassName = $ShippingFulfillmentType->className;
#----------------------------------------------------------------------------------------
# description  Returns the class name, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $ClassName | class name | String
#========================================================================================
sub className  { 'ShippingFulfillmentType' }

#========================================================================================
# function     dbPackage
# state        public
#----------------------------------------------------------------------------------------
# syntax       $DbPackage = $ShippingFulfillmentType->dbPackage;
#----------------------------------------------------------------------------------------
# description  Returns the database package, used for DAL access.
#----------------------------------------------------------------------------------------
# return       $DbPackage | database package name | String
#========================================================================================
sub dbPackage { __PACKAGE__ }

#========================================================================================
# function     insert
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentType = DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentType->insert($hInfo);
# example      my $Class = LoadClassByAlias('ShippingFulfillmentType');
#               my $ShippingFulfillmentType = $Class->insertObject($hInfo);
#----------------------------------------------------------------------------------------
# description  Insert this info as object, afterwards the hook 'OBJ_InsertShippingFulfillmentType' is triggered.
#----------------------------------------------------------------------------------------
# input        $hInfo | attributes of new object | ref.hash
# return       $ShippingFulfillmentType | new object | object
# hook         OBJ_InsertShippingFulfillmentType | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentTypeID - shipping fulfillment type object id - int
#               </ul>
#========================================================================================
sub insert {
    my $class = shift;
    my ($hInfo) = @_;

    my $self = $class->SUPER::insert($hInfo);

    TriggerHook('OBJ_InsertShippingFulfillmentType', {'Object' => $self, 'ShippingFulfillmentTypeID' => $self->id });
    delete $hInfo->{$_} foreach qw (
        SizeUnit
        WeightUnit
        MethodClass
        ShippingTransactionClass
        ShippingLabelClass
        CanBatchAction
        IntegrationType
        Logging
        WebserviceURL
        WebserviceTestURL
        AdminURL
        AdminTestURL
        RegistrationURL
        ExportDriver
        ImportHandler
        ExportFilename
        ProviderMethodClass
    );

    return $self;
}

#========================================================================================
# function     delete
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentType->delete;
#----------------------------------------------------------------------------------------
# description  Deletes this object, therefore the hook 'OBJ_DeleteShippingFulfillmentType' is triggered.
#----------------------------------------------------------------------------------------
# hook         OBJ_DeleteShippingFulfillmentType | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentTypeID - shipping fulfillment type object id - int
#               </ul>
#========================================================================================
sub delete {
    my $self = shift;

    TriggerHook('OBJ_DeleteShippingFulfillmentType', {'Object' => $self, 'ShippingFulfillmentTypeID' => $self->id });
    LoadClassByAlias('ShippingFulfillmentMethod')->deleteObjects('ByShippingFulfillmentTypeID', [$self->id]);
    LoadClassByAlias('ShippingProviderMethod')->deleteObjects('ByShippingFulfillmentTypeID', [$self->id]);
    LoadClassByAlias('ShippingFulfillmentOption')->deleteObjects('ByShippingFulfillmentTypeID', [$self->id]);
    LoadClassByAlias('ShippingFulfillmentProduct')->deleteObjects('ByShippingFulfillmentTypeID', [$self->id]);
    LoadClassByAlias('ShippingTransaction')->deleteObjects('ByShippingFulfillmentTypeID', [$self->id]);
    $self->SUPER::delete;
    return;
}

#========================================================================================
# function     set
# state        private
#----------------------------------------------------------------------------------------
# syntax       $ShippingFulfillmentType->set($hValues, $LanguageID);
#----------------------------------------------------------------------------------------
# description  This function triggers hooks 'OBJ_BeforeUpdateShippingFulfillmentType' and 'OBJ_AfterUpdateShippingFulfillmentType'.
#----------------------------------------------------------------------------------------
# input        $hValues | attribute name => value hash | ref.hash
# input        $LanguageID | (optional; required for localized attributes) language id | int
#----------------------------------------------------------------------------------------
# hook         OBJ_BeforeUpdateShippingFulfillmentType | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentTypeID - shipping fulfillment type object id - int
#                 <li>Values        - new values for object - ref.hash
#                 <li>LanguageID    - values of this language will be updated - int
#               </ul>
# hook         OBJ_AfterUpdateShippingFulfillmentType | hook parameter keys :
#               <ul>
#                 <li>Object        - this object          - object
#                 <li>ShippingFulfillmentTypeID - shipping fulfillment type object id - int
#                 <li>Values        - updated values       - ref.hash
#                 <li>LanguageID    - values updated of this language - int
#               </ul>
#========================================================================================
sub set {
    my $self = shift;
    my ($hValues, $LanguageID) = @_;

    TriggerHook('OBJ_BeforeUpdateShippingFulfillmentType', {'Object'=>$self, 'Values'=> $hValues, 'ShippingFulfillmentTypeID' => $self->id, 'LanguageID' => $LanguageID});
    $self->SUPER::set($hValues, $LanguageID);
    TriggerHook('OBJ_AfterUpdateShippingFulfillmentType',  {'Object'=>$self, 'Values'=> $hValues, 'ShippingFulfillmentTypeID' => $self->id, 'LanguageID' => $LanguageID});
    return;
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::ExternalShipping::API::Object::ShippingFulfillmentType->default( $Parent, $hInfo );
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $class = shift;
    my ($Parent, $hInfo) = @_;
    return {
        %{$class->SUPER::default($Parent, $hInfo)},
        'CanBatchAction' => 0,
        'IntegrationType' => '',
        'Logging' => 0,
    };
}

#========================================================================================
# function     importXML
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Object = DE_EPAGES::Object::API::Object::Object->importXML($Handler, $hElement, $Class, $hAttributes);
#----------------------------------------------------------------------------------------
# description  Imports an object from an XML file. This function is called at the XML
#               start tag. The end tag triggers the <function importXMLEnd> function.
#----------------------------------------------------------------------------------------
# input        $Handler | xml import handler with support function addObject | object
# input        $hElement | XML element structure, see XML::SAX | ref.hash
# input        $Class | class object; the new object will be an instance of this class | object
# input        $hAttributes | attributes hash | ref.hash
# return       $Object | the imported object | object
#========================================================================================
sub importXML {
    my $self = shift;
    my ($Handler, $hElement, $Class, $hAttributes) = @_;

    return $self->SUPER::importXML($Handler, $hElement, $Class, $hAttributes) if $Handler->topID('IsReference');

    foreach my $ClassName (qw(ShippingLabelClass MethodClass ShippingTransactionClass ProviderMethodClass)){
        $hAttributes->{$ClassName} = LoadClassByAlias($hAttributes->{$ClassName}) if defined $hAttributes->{$ClassName};
    }

    return $self->SUPER::importXML($Handler, $hElement, $Class, $hAttributes);
}

#========================================================================================
# function     newShippingFulfillmentMethod
# state        public
#----------------------------------------------------------------------------------------
# syntax       newShippingFulfillmentMethod( $Shop, $hInfo );
#----------------------------------------------------------------------------------------
# description  Creates a new ShippingFulfillmentMethod and sets defaults.
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $hInfo | attributes to set for newly generated method | ref.hash
# return       $Method | shipping fulfillment method | object
#========================================================================================
sub newShippingFulfillmentMethod {
    my $self = shift;
    my ($Shop, $hInfo) = @_;
    $hInfo->{'Parent'} = GetShippingFulfillmentMethodsFolder($Shop);
    $hInfo->{'ShippingFulfillmentType'} = $self;
    $hInfo->{'Alias'} = $hInfo->{'Parent'}->nextAlias('ShippingFulfillmentMethod%4d') unless isValid( $hInfo->{'Alias'} ) ;
    my $MethodClass = $self->get('MethodClass');
    $MethodClass = LoadClassByAlias('ShippingFulfillmentMethod') unless defined $MethodClass;
    return $MethodClass->insertObject($hInfo);
}

#========================================================================================
# function     newShippingProviderMethod
# state        public
#----------------------------------------------------------------------------------------
# syntax       newShippingProviderMethod( $Shop, $hInfo );
#----------------------------------------------------------------------------------------
# description  Creates a new ShippingProviderMethod and sets defaults.
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# input        $hInfo | attributes to set for newly generated method | ref.hash
# return       $Method | shipping provider method | object
#========================================================================================
sub newShippingProviderMethod {
    my $self = shift;
    my ($Shop, $hInfo) = @_;
    $hInfo->{'Parent'} = GetShippingProviderMethodsFolder($Shop);
    $hInfo->{'ShippingFulfillmentType'} = $self;
    $hInfo->{'Alias'} = $hInfo->{'Parent'}->nextAlias('ShippingProviderMethod%4d') unless isValid( $hInfo->{'Alias'} ) ;
    my $ProviderMethodClass = $self->get('ProviderMethodClass');
    $ProviderMethodClass = LoadClassByAlias('ShippingProviderMethod') unless defined $ProviderMethodClass;
    return $ProviderMethodClass->insertObject($hInfo);
}

#========================================================================================
# function     isDeprecated
# state        public
#----------------------------------------------------------------------------------------
# syntax       isDeprecated( $Shop);
#----------------------------------------------------------------------------------------
# description  Checks if ShippingFulfillment type was marked as deprecated
#----------------------------------------------------------------------------------------
# input        $Shop | shop | object
# return       boolean
#========================================================================================

sub isDeprecated {
    my $self = shift;
    my ($Shop) = @_;

    my $DeprecationDate;
    if ($Shop->get('IsUnityRegistered')) {
        $DeprecationDate = $self->get('DeprecationDateUnity');
    }
    else {
        $DeprecationDate = $self->get('DeprecationDate');
    }

    my $TimeZone = GetLocalTimeZone();

    my $DateTimeToday = DateTime->today('time_zone'=> $TimeZone);

    if ((!defined $DeprecationDate ||
            (DateTime->compare($DeprecationDate, $DateTimeToday) > 0))) {
        return 0;
    }
    return 1;
}

1;
