use Getopt::Long;
use DE_EPAGES::Cartridge::API::Table::Cartridge qw ( ExistsCartridgeByPackage );
use DE_EPAGES::Core::API::File qw ( ExistsFile  );
use DE_EPAGES::Core::API::Script qw( RunScript );
use DE_EPAGES::Database::API::Connection qw( RunOnStore );
use DE_EPAGES::Object::API::Factory qw ( LoadObjectByPath JoinObjectPath LoadClassByAlias );
use Data::Dumper;
use DateTime;
use feature 'switch';

use strict;
use warnings;
my $NoOutput;
my $IsDryRun;
my $StoreName;

sub Main {
    local $| = 1;

    my $Help;
    my $Password;
    my $ListFile;
    my $ShippingMethod;
    GetOptions(
        'help'          => \$Help,
        'nooutput'      => \$NoOutput,
        'dryrun'        => \$IsDryRun,
        'passwd=s'      => \$Password,
        'storename=s'   => \$StoreName,
        'list=s'        => \$ListFile,
        'shipping=s'     => \$ShippingMethod,
    );
    usage() if $Help;
    usage() unless ( defined($ListFile) || defined($ShippingMethod) );
    usage() unless defined $StoreName;

    my @Lines = ();
    if (defined $ListFile) {
        fileNotFound($ListFile) unless (ExistsFile($ListFile));

        open my $handle, '<', $ListFile or die "Unable to open $ListFile\n";
        chomp(@Lines = <$handle>);
        close $handle or die "Unable to close $ListFile\n";
    }

    if (defined $ShippingMethod) {
        push(@Lines, $ShippingMethod);
    }

    my $time1 = time;

    RunOnStore( Store => $StoreName, DBPassword => $Password, Sub => sub {
        foreach my $DeprecatedExternalShipping (@Lines) {
            my ($ShippingFulfillmentTypeName, $CartridgeName) = split(/:/, $DeprecatedExternalShipping);
            unless (ExistsCartridgeByPackage('DE_EPAGES::' . $CartridgeName)) {
                print "Store \"$StoreName\" : Cartridge \"DE_EPAGES::$CartridgeName\" is not installed (SKIP).\n" unless $NoOutput;
                next;
            }
            my $ShippingFulFillmentTypeObject = LoadObjectByPath(JoinObjectPath(['ShippingFulfillmentTypes', $ShippingFulfillmentTypeName]));

            my $rAreValidSettings;

            given ($ShippingFulfillmentTypeName) {
                when ('Packlink') {
                    $rAreValidSettings = \&AreValidSettingsForPacklink;
                }

                default {
                    print "Store \"$StoreName\" : Shipping fulfillment type \"$ShippingFulfillmentTypeName\" is not covered (SKIP).\n" unless $NoOutput;
                    next;
                }
            }
            my @aShippingMethodsWithDeprecatedType = FindDeprecatedShippingMethods($ShippingFulFillmentTypeObject);
            RemoveNotSetShippingMethods($rAreValidSettings, @aShippingMethodsWithDeprecatedType);
        }
    });

    my $time2 = time;
    my $seconds = $time2-$time1;
    my $logmsg = sprintf ('need %.3f seconds', $seconds);
    print $logmsg."\n" unless $NoOutput;
}

sub FindDeprecatedShippingMethods {
    my ($ShippingFulfillmentTypeObject) = @_;

    my $hShippingFulfillmentTypeProperties = $ShippingFulfillmentTypeObject->get([
        'DeprecationDate',
        'DeprecationDateUnity',
        'ID',
        'Alias',
        'ProviderMethodClass',
        'MethodClass'
    ]);

    if (defined $hShippingFulfillmentTypeProperties->{'MethodClass'}) {
        $hShippingFulfillmentTypeProperties->{'MethodClassAlias'} = $hShippingFulfillmentTypeProperties->{'MethodClass'}->alias();
    }

    if (defined $hShippingFulfillmentTypeProperties->{'ProviderMethodClass'}) {
        $hShippingFulfillmentTypeProperties->{'ProviderMethodClassAlias'} = $hShippingFulfillmentTypeProperties->{'ProviderMethodClass'}->alias();
    }

    my $ShippingFulfillmentTypeID = $hShippingFulfillmentTypeProperties->{'ID'};
    my $ShippingFulfillmentTypeAlias = $hShippingFulfillmentTypeProperties->{'Alias'};
    print "ShippingFulfillmentType Object ID: $ShippingFulfillmentTypeID\n" unless $NoOutput;

    my $ShippingMethodClass = LoadClassByAlias('ShippingMethod'); #/Classes/ShippingMethod
    my $aShippingMethodObjects = $ShippingMethodClass->getObjects();

    my @aShippingMethodsWithDeprecatedType;
    foreach my $ShippingMethodObject (@{ $aShippingMethodObjects }) {
        if (IsShippingDeprecated($ShippingMethodObject, $hShippingFulfillmentTypeProperties)) {
            push(@aShippingMethodsWithDeprecatedType, $ShippingMethodObject);
        }
    }

    my $objectCount = scalar @aShippingMethodsWithDeprecatedType;
    print "Found $objectCount Objects in Store \n" unless $NoOutput;
    print "Store \"$StoreName\" : Process Shipping methods with \"$ShippingFulfillmentTypeAlias\" provider or fulfillment\n" unless $NoOutput;

    return @aShippingMethodsWithDeprecatedType;
}

sub IsShippingDeprecated {
    my ($ShippingMethodObject, $hShippingFulfillmentTypeProperties) = @_;

    my $Shop = $ShippingMethodObject->get('Site');

    my $DeprecationDate = $hShippingFulfillmentTypeProperties->{'DeprecationDate'};

    if ($Shop->get('IsUnityRegistered')) {
        $DeprecationDate = $hShippingFulfillmentTypeProperties->{'DeprecationDateUnity'};
    }

    unless (defined($DeprecationDate) && (DateTime->compare($DeprecationDate, DateTime->now()) < 0)) {
        return 0;
    }

    my $hShippingFulfillmentMethodOrProvider = $ShippingMethodObject->get([ 'SelectedShippingFulfillmentMethod', 'SelectedShippingProviderMethod' ]);

    if (defined $hShippingFulfillmentMethodOrProvider->{'SelectedShippingFulfillmentMethod'} &&
        defined $hShippingFulfillmentTypeProperties->{'MethodClassAlias'} && 
        $hShippingFulfillmentMethodOrProvider->{'SelectedShippingFulfillmentMethod'}->instanceOf($hShippingFulfillmentTypeProperties->{'MethodClassAlias'}))
    {
        return 1;
    }

    if (defined $hShippingFulfillmentMethodOrProvider->{'SelectedShippingProviderMethod'} &&
        defined $hShippingFulfillmentTypeProperties->{'ProviderMethodClassAlias'} &&
        $hShippingFulfillmentMethodOrProvider->{'SelectedShippingProviderMethod'}->instanceOf($hShippingFulfillmentTypeProperties->{'ProviderMethodClassAlias'}))
    {
        return 1;
    }

    return 0;
}


sub RemoveNotSetShippingMethods {
    my ($rAreValidSettings, @aShippingMethodsWithDeprecatedType) = @_;

    foreach my $DeprecatedShippingMethod (@aShippingMethodsWithDeprecatedType) {
        unless ($rAreValidSettings->($DeprecatedShippingMethod)) {
            if ($IsDryRun) {
                print "\t\tFOUND '" . $DeprecatedShippingMethod->get('Path') . "' shipping method in Shop ". $DeprecatedShippingMethod->get('Site')->get('Alias') . "\n" unless $NoOutput;
            } else {
                $DeprecatedShippingMethod->delete();
                print "\t\tDELETED '" . $DeprecatedShippingMethod->get('Path') . "' shipping method in Shop ". $DeprecatedShippingMethod->get('Site')->get('Alias') . "\n" unless $NoOutput;
            }
        }
    }
}
#=======================================================
# ShippingMethod checkers
#=======================================================
sub AreValidSettingsForPacklink {
    my ($ShippingMethod) = @_;

    my $hShippingInfo = $ShippingMethod->get([ 'SelectedShippingFulfillmentMethod', 'SelectedShippingProviderMethod' ]);

    my $ShippingFulfillmentMethodOrProvider = $hShippingInfo->{'SelectedShippingFulfillmentMethod'};

    if (defined $hShippingInfo->{'SelectedShippingProviderMethod'}) {
        $ShippingFulfillmentMethodOrProvider = $hShippingInfo->{'SelectedShippingProviderMethod'};
    }

    return $ShippingFulfillmentMethodOrProvider->isUsable();
}

#========================================================
sub usage {
    print <<END_USAGE;
Usage:
    perl $0 [options] flags objects

options:
    -passwd     database user password                     ()
    -storename  name of store                              ()
    -list       file with deprecated shippingmethods in the form:
                <shipping_name>:<cartridge_name>       ()
    -shipping    single deprecated shipping method in the form:
                <shipping_name>:<cartridge_name>       ()

flags:
    -help       show the command line options
    -nooutput   do not print any output
    -dryrun     do not remove any Shipping method

Examples:
    perl $0 -storename Store -list /var/tmp/payment_cartridges.list
    perl $0 -storename Store -shipping Acceptance:Ogone -dryrun

END_USAGE
    exit 2;
}

sub fileNotFound {
    my ($path) = @_;

    print "file \"$path\" not found\n";
    exit 2;
}

RunScript(
    'Sub' => \&Main,
);
