#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::API::ProductBatchService
# state        public
#----------------------------------------------------------------------------------------
# description  GoogleSmartShoppingScriptTagService class
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::API::ProductBatchService;

use DE_EPAGES::Core::API::Error qw ( Error ExistsError GetError );
use DE_EPAGES::Core::API::Log qw( LogError );
use DE_EPAGES::Core::API::Warning qw (Warning);

use DE_EPAGES::GoogleSmartShopping::API::GoogleApiService;
use DE_EPAGES::GoogleSmartShopping::API::Tools qw ( LogGoogleSmartShopping GetErrorCode );

use DE_EPAGES::GoogleSmartShopping::API::Constants qw (
    NUMBER_PRODUCTS_REQUEST_THRESHOLD
    ERROR_REQUEST_RATE_TOO_HIGH
    ERROR_DAILY_LIMIT_EXCEEDED
    ERROR_ACCESS_DENIED
);

use constant GOOGLE_MERCHANT_CENTER_SUPPORT_FORM_URL => 'https://support.google.com/merchants/contact/additional_items';

use strict;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::GoogleSmartShopping::API::ProductBatchService->new( $Shop );
#----------------------------------------------------------------------------------------
# description  constructor
#----------------------------------------------------------------------------------------
# return       $self | self | DE_EPAGES::GoogleSmartShopping::API::ProductBatchService
#========================================================================================
sub new {
    my $class = shift;
    my ($Shop)  = @_;

    my $self = bless {
        'Shop' => $Shop,
        'Products' => [],
    }, $class;

    return $self;
}

#========================================================================================
# function     ExportItem
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductBatchService->ExportItem( $hProduct );
#----------------------------------------------------------------------------------------
# description  Receives each product data and saves it on $self->{'Products'}
#----------------------------------------------------------------------------------------
# input        $hProduct | a product data for the export | refhash
#========================================================================================
sub ExportItem {
    my $self = shift;
    my ($hProduct)  = @_;

    push(@{$self->{'Products'}}, {
        'method' => 'insert',
        'product' => $hProduct,
    });

    if (scalar @{$self->{'Products'}} >= NUMBER_PRODUCTS_REQUEST_THRESHOLD) {
        $self->_sendBatchRequest();
    }

    return;
}

#========================================================================================
# function     OnStartBatch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductBatchService->OnStartBatch();
#----------------------------------------------------------------------------------------
# description  OnStartBatch gets called before the start of the iteration of products
#               It loads the GoogleApiService necessary to make the insert/delete requests
#========================================================================================
sub OnStartBatch {
    my $self = shift;
 
    $self->{'GoogleApiService'} = DE_EPAGES::GoogleSmartShopping::API::GoogleApiService->new({Shop => $self->{'Shop'}});
    $self->{'GoogleApiService'}->loadGoogleSettings();

    $self->{'MerchantCenterAccountID'} = $self->{'GoogleApiService'}->{'GoogleSmartShoppingConfig'}->get('MerchantCenterAccountID');

    return;
}

#========================================================================================
# function     OnEndBatch
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductBatchService->OnEndBatch();
#----------------------------------------------------------------------------------------
# description  OnEndBatch gets called at the end of the iteration of products
#               It triggers _sendBatchRequest if products are waiting in queue
#========================================================================================
sub OnEndBatch {
    my $self = shift;
 
    if (scalar @{$self->{'Products'}}) {
        $self->_sendBatchRequest();
    }

    return;
}

#========================================================================================
# function     setBatchOfProducts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductBatchService->setBatchOfProducts($aProducts);
#----------------------------------------------------------------------------------------
# description  setBatchOfProducts sets batch of products to process into batch service object.
#----------------------------------------------------------------------------------------
# input        $aProducts | array of product IDs to delete | array.ref
#========================================================================================
sub setBatchOfProducts {
    my $self = shift;
    my ($aProducts) = @_;

    $self->{'Products'} = $aProducts;

    return;
}

#========================================================================================
# function     deleteProducts
# state        public
#----------------------------------------------------------------------------------------
# syntax       $ProductBatchService->deleteProducts($aProducts);
#----------------------------------------------------------------------------------------
# description  It contacts Google Ads services to delete provided list of products.
#----------------------------------------------------------------------------------------
# input        $aProducts | array of product IDs to delete | array.ref
#========================================================================================
sub deleteProducts {
    my $self = shift;
    my ($aProducts) = @_;

    my @ProductsDeletion = map { 'method' => 'delete', 'productId' => $_ }, @$aProducts;
    $self->setBatchOfProducts(\@ProductsDeletion);
    $self->_sendBatchRequest();

    return;
}

sub _sendBatchRequest {
    my $self = shift;

    my $hProductBatchPayload = { 'entries' => $self->{'Products'} };

    my $hResponse;
    eval {
        $hResponse = $self->{'GoogleApiService'}->callProductsBatch($hProductBatchPayload);
    };
    if (ExistsError()) {
        my $Error = $self->_GetError();
        LogGoogleSmartShopping('sendBatchRequest failed', {
            'Error' => $Error->message(),
            'Response' => $hResponse,
        });
        my $ErrorCode = $self->_GetErrorCode($Error);
        if ($ErrorCode ~~ [ERROR_REQUEST_RATE_TOO_HIGH, ERROR_DAILY_LIMIT_EXCEEDED]  ) {
            $self->_LogGoogleApiQuotaError($Error);
            $self->_Warning($ErrorCode, { 'Error' => $Error } );
        } elsif ($ErrorCode ~~ [ERROR_ACCESS_DENIED]  ) {
            LogError($Error);
            $self->_Warning($ErrorCode, { 'Error' => $Error } );
        } else {
            $self->_Warning($Error);
        }
    }

    $self->{'Products'} = [];

    return;
}


#----------------------------------------------------------------------------------------
# Helpers
#----------------------------------------------------------------------------------------
sub _GetError {
    my $self = shift;
    return GetError();
}

sub _GetErrorCode {
    my $self = shift;
    my ($Error) = @_;
    return GetErrorCode($Error);
}

sub _Warning {
    my $self = shift;
    my ($Code, $ErrorParams) = @_;

    Warning($Code, $ErrorParams);
}

sub _LogGoogleApiQuotaError{
    my $self = shift;
    my ($GoogleError) = @_;

    eval { Error('GoogleApiQuotaError', {
        'GoogleApiQuotaError' => $GoogleError,
        'SupportLink'       => GOOGLE_MERCHANT_CENTER_SUPPORT_FORM_URL
    }); };
    if (ExistsError()) {
        my $Error = GetError();
        LogError($Error);
    }
}

1;
