#========================================================================================
# package      DE_EPAGES::GoogleSmartShopping::t::unit::ProductBatchService
# base         DE_EPAGES::Core::API::Test
# state        public
#----------------------------------------------------------------------------------------
# description  Unit test for API::ProductBatchService
#========================================================================================
package DE_EPAGES::GoogleSmartShopping::t::unit::ProductBatchService;
use base DE_EPAGES::Core::API::Test;


use DE_EPAGES::GoogleSmartShopping::API::GoogleApiService;

use DE_EPAGES::Test::Mock::MockifyEpagesObject;
use Test::Mockify;
use Test::Mockify::Sut;
use Test::Exception;
use Test::MockObject::Extends;
use Test::Mockify::Verify qw ( WasCalled GetParametersFromMockifyCall );
use Test::Mockify::Matcher qw ( String HashRef ArrayRef);
use JSON::MaybeXS qw ( encode_json );

use Test::More;

use strict;

use constant LOCALE_ID => 'en_GB';
use constant DAILY_LIMIT_EXCEEDED_API_ERROR => 'quota/daily_limit_exceeded' ;
use constant DAILY_LIMIT_EXCEEDED_EXPORT_WARNING => 'daily_limit_exceeded' ;
use constant ACCOUNT_ACCESS_DENIED_API_ERROR => 'auth/account_access_denied' ;
use constant ACCOUNT_ACCESS_DENIED_EXPORT_WARNING => 'access_denied' ;
use constant EXPECTED_SUPPORT_FORM_URL => 'https://support.google.com/merchants/contact/additional_items';

use DE_EPAGES::GoogleSmartShopping::API::Constants qw ( NUMBER_PRODUCTS_REQUEST_THRESHOLD );

use constant PRODUCT_EXPORT_DATA => {
    "link" => "https://demoldhgst4tgga.vm-hsilva.intern.epages.de/p/unisex-sneakers-brombeer",
    "isBundle" => "false",
    "productTypes" => ["Schuhe"],
    "identifierExists" => "false",
    "description" => "The modern alternative to the classic lace-up sneakers. An easy-on design with a subtle print and elastic inserts for a comfortable and flexible fit. The perfect update to your wardrobe! Animal print in firm textile Off-white rubber sole Fylly leather inside Cushioned insole Removable insole",
    "availability" => "in stock",
    "offerId" => "56-sneakers_de_DE",
    "title" => "Unisex sneakers - brombeer",
    "imageLink" => "http://demoldhgst4tgga.vm-hsilva.intern.epages.de/WebRoot/Store/Shops/demoldhgst4tgga/Products/56-sneakers/sneaker-classic-brombeer.jpg",
    "price" => {
        "currency" => "EUR",
        "value" => "79.90"
    },
    "shipping" => [
        {
            "country" => "DE",
            "price" => {
                "currency" => "EUR",
                "value" => "7.00"
            }
         }
    ],
    "condition" => "new",
};

use constant PRODUCT_ID_TO_DELETE_1 => 'online:de:DE:umbrella_de_DE';
use constant PRODUCT_ID_TO_DELETE_2 => 'online:de:DE:watch_de_DE';
use constant PRODUCT_ID_TO_DELETE_3 => 'online:de:DE:purse_de_DE';
use constant PRODUCT_DELETE_IDS => [ PRODUCT_ID_TO_DELETE_1, PRODUCT_ID_TO_DELETE_2, PRODUCT_ID_TO_DELETE_3 ];

sub test {
    my $self = shift;

    $self->test_OnStartBatch();
    $self->test_ExportItem_Less_than_Threshold();
    $self->test_ExportItem_Bigger_than_Threshold();
    $self->test_OnEndBatch_empty_products();
    $self->test_OnEndBatch_existing_products();
    $self->test_deleteProducts();
    $self->test_ExportItem_throws_warning_when_daily_limit_exceeded();
    $self->test_ExportItem_throws_warning_when_auth_access_denied_error();

    done_testing();
}

sub test_OnStartBatch {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];
    
    my $MockedShop = $self->_createShop();

    my $MockedGoogleConfig = $self->_createGoogleConfig();
    my $MockedGoogleApiService = $self->_createGoogleApiService($MockedShop, $MockedGoogleConfig);
    my $SUT = $self->_createSUT($MockedShop, $MockedGoogleApiService);

    lives_ok(
        sub {
            $SUT->OnStartBatch();
        },
        "$SubTestName - OnStartBatch can be called without errors."
    );

    ok(
        WasCalled($MockedGoogleApiService, 'loadGoogleSettings'),
        $SubTestName . ' - Verify that loadGoogleSettings was called.'
    );

    ok(
        WasCalled($MockedGoogleConfig, 'get'),
        $SubTestName . ' - Verify that get MerchantCenterAccountID was called.'
    );

    is_deeply(
        $SUT->{'GoogleApiService'},
        $MockedGoogleApiService,
        $SubTestName . ' - Verify that GoogleApiService was set up correctly'
    );

    is(
        $SUT->{'MerchantCenterAccountID'},
        12345,
        $SubTestName . ' - Verify that MerchantCenterAccountID was set up correctly'
    );

    is_deeply($SUT->{'Products'}, [], $SubTestName . ' - Verify that Products array is empty');
}

sub test_ExportItem_Less_than_Threshold{
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];
    
    my $MockedShop = $self->_createShop();

    my $SUT = $self->_createSUT($MockedShop);

    $SUT->OnStartBatch();

    lives_ok(
        sub {
            $SUT->ExportItem(PRODUCT_EXPORT_DATA);
        },
        "$SubTestName - ExportItem can be called without errors."
    );

    is(@{$SUT->{'Products'}}, 1, $SubTestName . ' - Verify that Products array has 1 product');

    ok(
        !WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was NOT called.'
    );

    lives_ok(
        sub {
            for (my $i = 0; $i < NUMBER_PRODUCTS_REQUEST_THRESHOLD - 2; $i++) {
                $SUT->ExportItem(PRODUCT_EXPORT_DATA);
            }
        },
        "$SubTestName - ExportItem can be called (NUMBER_PRODUCTS_REQUEST_THRESHOLD - 2) without errors."
    );

    ok(
        !WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was NOT called.'
    );

    is(scalar @{$SUT->{'Products'}}, NUMBER_PRODUCTS_REQUEST_THRESHOLD - 1, $SubTestName . ' - Verify that Products array is (NUMBER_PRODUCTS_REQUEST_THRESHOLD - 1)');
}

sub test_ExportItem_Bigger_than_Threshold{
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedGoogleConfig = $self->_createGoogleConfig();
    my $MockedShop = $self->_createShop();
    my $MockedGoogleApiService = $self->_createGoogleApiService($MockedShop, $MockedGoogleConfig);

    my $SUT = $self->_createSUT($MockedShop, $MockedGoogleApiService);

    $SUT->OnStartBatch();

    lives_ok(
        sub {
            for (my $i = 0; $i < NUMBER_PRODUCTS_REQUEST_THRESHOLD; $i++) {
                $SUT->ExportItem(PRODUCT_EXPORT_DATA);
            }
        },
        "$SubTestName - ExportItem can be called n amount of times without errors."
    );


    ok(
        WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was called.'
    );

    ok(
        WasCalled($SUT->{'GoogleApiService'}, 'callProductsBatch'),
        $SubTestName . ' - Verify that callProductsBatch was called.'
    );

    my $aCallProductsBatchParams = GetParametersFromMockifyCall($SUT->{'GoogleApiService'}, 'callProductsBatch');

    is_deeply(
        $aCallProductsBatchParams->[0]->{'entries'}->[0]->{'product'},
        PRODUCT_EXPORT_DATA,
        $SubTestName . ' - Verify that callProductsBatch has correct parameters on first product'
    );

    is_deeply(
        $aCallProductsBatchParams->[0]->{'entries'}->[NUMBER_PRODUCTS_REQUEST_THRESHOLD - 1]->{'product'},
        PRODUCT_EXPORT_DATA,
        $SubTestName . ' - Verify that callProductsBatch has correct parameters on last product'
    );

    is_deeply($SUT->{'Products'}, [], $SubTestName . ' - Verify that Products array is emptied after upload');
}

sub test_OnEndBatch_empty_products {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];
    
    my $MockedShop = $self->_createShop();

    my $SUT = $self->_createSUT($MockedShop);

    $SUT->OnStartBatch();

    is_deeply($SUT->{'Products'}, [], $SubTestName . ' - Verify that Products array is empty');

    lives_ok(
        sub {
            $SUT->OnEndBatch();
        },
        "$SubTestName - OnEndBatch can be called without errors."
    );

    ok(
        !WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was NOT called when empty Product array.'
    );

}

sub test_OnEndBatch_existing_products {
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];
    
    my $MockedShop = $self->_createShop();

    my $SUT = $self->_createSUT($MockedShop);

    $SUT->OnStartBatch();

    $SUT->ExportItem(PRODUCT_EXPORT_DATA);

    is(@{$SUT->{'Products'}}, 1, $SubTestName . ' - Verify that Products array has 1 product');

    lives_ok(
        sub {
            $SUT->OnEndBatch();
        },
        "$SubTestName - OnEndBatch can be called without errors."
    );


    ok(
        WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was called.'
    );

    ok(
        WasCalled($SUT->{'GoogleApiService'}, 'callProductsBatch'),
        $SubTestName . ' - Verify that callProductsBatch was called.'
    );

    my $aCallProductsBatchParams = GetParametersFromMockifyCall($SUT->{'GoogleApiService'}, 'callProductsBatch');

    is_deeply(
        $aCallProductsBatchParams->[0]->{'entries'}->[0]->{'product'},
        PRODUCT_EXPORT_DATA,
        $SubTestName . ' - Verify that callProductsBatch has correct parameters on first product'
    );

    is_deeply($SUT->{'Products'}, [], $SubTestName . ' - Verify that Products array is emptied after upload');
}

sub test_deleteProducts {
    my $self = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedGoogleConfig = $self->_createGoogleConfig();
    my $MockedShop = $self->_createShop();
    my $MockedGoogleApiService = $self->_createGoogleApiService($MockedShop, $MockedGoogleConfig);

    my $SUT = $self->_createSUT($MockedShop, $MockedGoogleApiService);
    $SUT->OnStartBatch();

    lives_ok(
        sub {
            $SUT->deleteProducts(PRODUCT_DELETE_IDS);
        },
        "$SubTestName - deleteProducts can be called without errors."
    );
    ok(
        WasCalled($SUT, 'setBatchOfProducts'),
        $SubTestName . ' - Verify that setBatchOfProducts was called.'
    );
    ok(
        WasCalled($SUT, '_sendBatchRequest'),
        $SubTestName . ' - Verify that _sendBatchRequest was called.'
    );
    ok(
        WasCalled($SUT->{'GoogleApiService'}, 'callProductsBatch'),
        $SubTestName . ' - Verify that callProductsBatch was called.'
    );

    my $aCallSetBatchOfProducts = GetParametersFromMockifyCall($SUT, 'setBatchOfProducts');
    is(
        $aCallSetBatchOfProducts->[0]->[0]->{'method'},
        'delete',
        $SubTestName . ' - Verify that delete method is set'
    );
    is(
        $aCallSetBatchOfProducts->[0]->[0]->{'productId'},
        PRODUCT_ID_TO_DELETE_1,
        $SubTestName . ' - Verify that product ids have been set through setBatchOfProducts call'
    );
    # my $aCallProductsBatchParams = GetParametersFromMockifyCall($SUT->{'GoogleApiService'}, 'callProductsBatch');
    # is(
    #     $aCallProductsBatchParams->[0]->{'entries'}->[0]->{'method'},
    #     'adsfadf',
    #     $SubTestName . ' - Verify that callProductsBatch has correct value on first method type'
    # );
    # is(
    #     $aCallProductsBatchParams->[0]->{'entries'}->[0]->{'productId'},
    #     'adsfadf',
    #     $SubTestName . ' - Verify that callProductsBatch has correct value on first productId'
    # );
}

sub test_ExportItem_throws_warning_when_daily_limit_exceeded{
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShop = $self->_createShop();
    my $MockedGoogleConfig = $self->_createGoogleConfig();
    my $MockedGoogleApiService = $self->_createGoogleApiService($MockedShop, $MockedGoogleConfig);

    my $hSUTParams = {
        'Error' => $self->_createGoogleApiError(DAILY_LIMIT_EXCEEDED_API_ERROR)
    };

    my $SUT = $self->_createSUT($MockedShop, $MockedGoogleApiService, $hSUTParams);

    $SUT->OnStartBatch();

    lives_ok(
        sub {
            for (my $i = 0; $i < NUMBER_PRODUCTS_REQUEST_THRESHOLD; $i++) {
                $SUT->ExportItem(PRODUCT_EXPORT_DATA);
            }
        },
        "$SubTestName - ExportItem can be called n amount of times without errors."
    );


    ok( WasCalled($SUT, '_sendBatchRequest'), $SubTestName . ' - Verify that _sendBatchRequest was called.');
    ok( WasCalled($SUT->{'GoogleApiService'}, 'callProductsBatch'), $SubTestName . ' - Verify that callProductsBatch was called.');
    ok( WasCalled($SUT, '_GetError'), $SubTestName . ' - Verify that _GetError was called.');
    ok( WasCalled($SUT, '_GetErrorCode'), $SubTestName . ' - Verify that _GetErrorCode was called.');

    ok(
        WasCalled($SUT, '_Warning'),
        $SubTestName . ' - Verify that _Warning was called.'
    );

    my $aParams = GetParametersFromMockifyCall($SUT, '_Warning');

    is_deeply($aParams->[0], DAILY_LIMIT_EXCEEDED_EXPORT_WARNING,
        "$SubTestName - Test if Warning was called with expected params.". $aParams->[0]);

    ok(
        WasCalled($SUT, '_LogGoogleApiQuotaError'),
        $SubTestName . ' - Verify that _LogGoogleApiQuotaError was called when quota error.');

}

sub test_ExportItem_throws_warning_when_auth_access_denied_error{
    my $self        = shift;
    my $SubTestName = ( caller(0) )[3];

    my $MockedShop = $self->_createShop();
    my $MockedGoogleConfig = $self->_createGoogleConfig();
    my $MockedGoogleApiService = $self->_createGoogleApiService($MockedShop, $MockedGoogleConfig);

    my $hSUTParams = {
        'Error' => $self->_createGoogleApiError(ACCOUNT_ACCESS_DENIED_API_ERROR)
    };

    my $SUT = $self->_createSUT($MockedShop, $MockedGoogleApiService, $hSUTParams);

    $SUT->OnStartBatch();

    lives_ok(
        sub {
            for (my $i = 0; $i < NUMBER_PRODUCTS_REQUEST_THRESHOLD; $i++) {
                $SUT->ExportItem(PRODUCT_EXPORT_DATA);
            }
        },
        "$SubTestName - ExportItem can be called n amount of times without errors."
    );


    ok( WasCalled($SUT, '_sendBatchRequest'), $SubTestName . ' - Verify that _sendBatchRequest was called.');
    ok( WasCalled($SUT->{'GoogleApiService'}, 'callProductsBatch'), $SubTestName . ' - Verify that callProductsBatch was called.');
    ok( WasCalled($SUT, '_GetError'), $SubTestName . ' - Verify that _GetError was called.');
    ok( WasCalled($SUT, '_GetErrorCode'), $SubTestName . ' - Verify that _GetErrorCode was called.');

    ok(!WasCalled($SUT, '_LogGoogleApiQuotaError'), $SubTestName . ' - Verify that _LogGoogleApiQuotaError was NOT called.');

    ok( WasCalled($SUT, '_Warning'), $SubTestName . ' - Verify that _Warning was called.');

    my $aParams = GetParametersFromMockifyCall($SUT, '_Warning');

    is_deeply($aParams->[0], ACCOUNT_ACCESS_DENIED_EXPORT_WARNING,
        "$SubTestName - Test if Warning was called with expected params.". $aParams->[0]);

}


#========================================================================================
# Mocking methods
#========================================================================================
sub _createSUT {
    my $self = shift;
    my ($MockedShop, $MockedGoogleApiService, $hSUTParams) = @_;

    my $MockifySUT = Test::Mockify::Sut->new('DE_EPAGES::GoogleSmartShopping::API::ProductBatchService', [$MockedShop]);

    $MockifySUT->mockImported('DE_EPAGES::GoogleSmartShopping::API::Tools', 'LogGoogleSmartShopping')
        ->whenAny()
        ->thenReturnUndef();

    if (defined($MockedGoogleApiService)) {
        $MockifySUT->mockStatic('DE_EPAGES::GoogleSmartShopping::API::GoogleApiService::new')
            ->when(String(), HashRef())
            ->thenReturn($MockedGoogleApiService);
    }

    if (defined($hSUTParams->{'Error'})) {
        $MockifySUT->mockImported('DE_EPAGES::Core::API::Error', 'ExistsError')
            ->whenAny()
            ->thenReturn(1);

        $MockifySUT->mock('_GetError')->whenAny()->thenReturn($hSUTParams->{'Error'});

    }
    
    $MockifySUT->spy('_sendBatchRequest')->when();

    $MockifySUT->spy('setBatchOfProducts')->when(ArrayRef());

    $MockifySUT->spy('_Warning')->whenAny();

    $MockifySUT->mock('_LogGoogleApiQuotaError')->whenAny()->thenReturnUndef();

    $MockifySUT->spy('_GetErrorCode')->whenAny();

    return $MockifySUT->getMockObject();
}

sub _createGoogleApiService {
    my $self = shift;
    my ($MockedShop, $MockedGoogleConfig) = @_;

    my $hServiceParams = {
        'Shop' => $MockedShop,
        'GoogleSmartShoppingConfig' => $MockedGoogleConfig,
    };

    my $Mockify = Test::Mockify->new('DE_EPAGES::GoogleSmartShopping::API::GoogleApiService', [$hServiceParams]);

    $Mockify->mock('loadGoogleSettings')->when()->thenReturnUndef();

    $Mockify->mock('callProductsBatch')->when(HashRef())->thenReturnUndef();

    my $MockedGoogleApiService = $Mockify->getMockObject();
    $MockedGoogleApiService->{'GoogleSmartShoppingConfig'} = $MockedGoogleConfig;

    return $MockedGoogleApiService;
}

sub _createGoogleConfig {
    my $self = shift;

    my $Mockify = Test::Mockify->new('DE_EPAGES::GoogleSmartShopping::API::Object::GoogleSmartShoppingConfig');

    $Mockify->mock('get')->when(String('MerchantCenterAccountID'))->thenReturn(12345);

    return $Mockify->getMockObject();
}

sub _createShop {
    my $self = shift;

    my $hAttributes = {
        'ID'            => undef,
        'LocaleID'      => LOCALE_ID,
        'LanguageID'    => 1,
        'NameOrAlias'   => 'ShopNameOrAlias',
    };
    my $MockifyShop = DE_EPAGES::Test::Mock::MockifyEpagesObject->new('DE_EPAGES::Shop::API::Object::Shop', $hAttributes);

    return $MockifyShop->getMockObject();
}

sub _createGoogleApiError {
    my $self = shift;
    my ( $Reason ) = @_;

    my $hResponse = {
        "errorId" => "google-api-error",
        "message" => "some error message",
        "details" => {
            "code" => 429,
            "errors" => [
                {
                    "domain" => "global",
                    "message" => "some message",
                    "reason" => $Reason
                }
            ],
            "message" => "Some error message 429 Too Many Requests"
        }
    };

    my $Error = DE_EPAGES::Core::API::Error->new(
        'Code' => 'Error',
        'Vars' => {
            'Status' => 429,
            'Response' => encode_json($hResponse)
        }
    );
    return $Error;

}


__PACKAGE__->runScript();

1;
