#==============================================================================
# package      DE_EPAGES::ProductPortal::UI::PortalConfig
# state        private
#==============================================================================
package DE_EPAGES::ProductPortal::UI::PortalConfig;
use base qw (DE_EPAGES::Presentation::UI::Object);

use strict;
use DE_EPAGES::Core::API::Error qw (Error ExistsError GetError);
use DE_EPAGES::Core::API::File qw (GetFileContent DeleteFile);
use DE_EPAGES::Core::API::Log qw (LogDebug);
use DE_EPAGES::Core::API::PerlTools qw (isValid trim ElementOf);
use DE_EPAGES::Core::API::Warning qw (BlockWarnings Warning);
use DE_EPAGES::Object::API::Factory qw (LoadClassByAlias);
use DE_EPAGES::Product::API::Object::PagerProducts;
use DE_EPAGES::Product::API::Product qw ( GetProductPrettyUnityUrlsRequest SaveProductPrettyUrlsResponse );
use DE_EPAGES::Product::API::Shop qw (GetProduct);
use DE_EPAGES::ProductPortal::API::Constants qw (MAX_WARNING_MESSAGES AMAZON GOOGLE_SMART_SHOPPING GOOGLE_BASE);
use DE_EPAGES::Search::API::Object::PagerSearch;
use DE_EPAGES::Search::API::SearchDefinition qw (GetSearchDefinition);
use HTTP::Status qw( HTTP_SERVICE_UNAVAILABLE );
use DE_EPAGES::ProductPortal::API::Portal qw ( LoadPortalSite GetPortalConfigFolder );
use JSON;
use utf8;
use Encode qw ( encode );

use constant MAX_REQUEST_BYTE_SIZE => 100000;

#==============================================================================
# Event handlers
#==============================================================================
sub attributeValues {
    my $self = shift;
    my ($Servlet, $Object, $Class) = @_;
    my $hhValues = $self->SUPER::attributeValues( $Servlet, $Object, $Class );

    if ($Object->instanceOf('PortalConfig') && $Object->get('PortalSite')->get('ExportsShippingPrice')) {
        my $ShippingMethod = $hhValues->{'Values'}->{'ShippingMethod'};
        my $hPortalConfig = {
            'CurrencyID' => $Object->get('CurrencyID'),
            'TaxModel'   => $Object->get('PortalSite')->get('TaxModel'),
        };

        # throw error if shop has no shipping methods or selected shipping method has no price
        $Servlet->form->addFormError({
            'Form'  => 'ExportPortalConfig',
            'Reason'=> 'NoShipping',
            'Name'  => 'ExportPortalConfig',
            'Value' => 'ShippingMethod'
        }) unless (defined($ShippingMethod) && $ShippingMethod->canAddToBasket($hPortalConfig));
    }
    return $hhValues;
}

#==============================================================================
sub UpdatePortalConfig {
    my $self = shift;
    my ($Servlet) = @_;

    my $Shop = $Servlet->object;
    my @ActivePortalConfigs = $self->_getAndFilterPortalConfigs($Shop, $Servlet->form->formValues);
    my @FormPortalIDs = $Servlet->form->value('PortalID');
    my @PortalIDsToAdd = grep { not ElementOf($_, map { $_->id } @ActivePortalConfigs) } @FormPortalIDs;
    my @PortalsToDelete = grep { not ElementOf($_->id, \@FormPortalIDs) } @ActivePortalConfigs;
    my @PortalsToRetain = grep { ElementOf($_->id, \@FormPortalIDs) } @ActivePortalConfigs;
    return unless (scalar @PortalIDsToAdd or scalar @PortalsToDelete);

    push (@PortalsToRetain, $self->_newPortalConfig($Servlet, $_)) foreach @PortalIDsToAdd;
    return unless (scalar @PortalsToDelete);

    if ($Servlet->user->get('DeleteConfirmation') and not $Servlet->form->value('ACCEPT')) {
        $Servlet->vars('DeleteObjects', [map { $_->tleHash } @PortalsToDelete]);
        $Servlet->vars('PortalConfigIDs', [map { { 'ObjectID' => $_->id } } @PortalsToRetain ]);
        $Servlet->vars('DELETE', 1);
        return;
    }

    foreach my $PortalForDeletion (@PortalsToDelete) {
        $self->_removeExportFile($PortalForDeletion);
        $PortalForDeletion->delete;
    }
}

#==============================================================================
sub _getAndFilterPortalConfigs {
    my $self = shift;
    my ($Shop, $FormValues) = @_;

    my $IsUnity = $Shop->get('IsUnityRegistered');
    my $aPortalConfigs = $Shop->get('PortalConfigs');
    my @PortalConfigs = grep { index($_->alias, GOOGLE_SMART_SHOPPING) == -1 } @$aPortalConfigs;
    if ($IsUnity and $FormValues->{'OnlyAmazon'}) {
        return grep { index($_->alias, AMAZON) == 0 } @PortalConfigs;
    } elsif ($IsUnity and $FormValues->{'OnlyGoogleShopping'}) {
        return grep { index($_->alias, GOOGLE_BASE) == 0 } @PortalConfigs;
    } elsif ($IsUnity) {
        return grep { index($_->alias, AMAZON) == -1 and index($_->alias, GOOGLE_BASE) == -1 } @PortalConfigs;
    } else {
        return @PortalConfigs;
    }
}

sub _removeExportFile {
    my $self = shift;
    my ($ActivePortalConfig) = @_;

    my $ExportFileName = $ActivePortalConfig->get('ExportFileName');
    my $Folder = $ActivePortalConfig->get('Portal')->get('ExportPublicPath');
    DeleteFile($Folder . '/' . $ExportFileName);

    return;
}

sub _newPortalConfig {
    my $self = shift;
    my ($Servlet, $PortalSiteID) = @_;

    my $Shop = $Servlet->object();

    my $PortalSite = LoadPortalSite($Shop, $PortalSiteID);
    return unless $PortalSite;

    my $hInfo = { 'Position' => 180 };

    my $PortalConfigFolder = GetPortalConfigFolder($Shop, $PortalSite, $hInfo);
    return unless $PortalConfigFolder;

    my $Portal = $PortalSite->get('Portal');
    my $Class = $Portal->get('PortalConfigClass');

    return $self->_createPortalConfig($Servlet, $PortalSite, $PortalConfigFolder, $Class);
}

sub _createPortalConfig {
    my $self = shift;
    my ($Servlet, $PortalSite, $Folder, $Class) = @_;

    return $self->NewObject($Servlet, $Folder, $Class, {
        'Alias'      => $PortalSite->alias(),
        'PortalSite' => $PortalSite,
    });
}

#==============================================================================
sub AddProductsFromTray {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    my $Tray = DE_EPAGES::Presentation::API::Object::Tray->load($Servlet->intSessionID);
    my $aProducts = $Tray->objectsByClass(LoadClassByAlias('Product'));

    my $Shop = $PortalConfig->getSite();
    foreach my $Product (@$aProducts) {
        Error('InvalidShop', {'a' => $Shop->id, 'b' => $Product->getSite()->id }) unless $Shop->id == $Product->getSite()->id;
    }
    my $aWarnings = BlockWarnings( sub {
        $PortalConfig->addProducts($aProducts);
    });
    my $Form = $Servlet->form;
    foreach my $warning (@$aWarnings) {
        $Form->mapFormError({
            'Form'  => 'AddProduct',
            'Name'  => 'ProductFromTray'
        }, $warning);
    }
    $Form->executeFormError;
    return;
}

#==============================================================================
sub RemoveProducts {
    my $self = shift;
    my ($Servlet) = @_;

    my $PortalConfig  = $Servlet->object;
    my $aProducts     = $PortalConfig->getSite()->loadObjects([$Servlet->form->value('ListObjectID')], 'Product');

    $Servlet->testPermission($_, 'Save') foreach @$aProducts;

    $PortalConfig->removeProducts($aProducts);
    return;
}

#==============================================================================
sub RemoveAllProducts {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    $PortalConfig->removeAllProducts;
    return;
}

#==============================================================================
sub ViewPagedChildren {
    my $self = shift;
    my ($Servlet) = @_;
    my $Pager = DE_EPAGES::Product::API::Object::PagerProducts->new({
        'Servlet' => $Servlet,
        'Object'  => $Servlet->object,
        'OrderBy' => 'Alias'
    });
    return $Pager->page;
}

#==============================================================================
sub SearchProducts {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    my $SearchDefinition = GetSearchDefinition(LoadClassByAlias('Product'), 'MBO-PortalConfig');
    my %Options = (
        'Servlet' => $Servlet,
        'Object'  => $PortalConfig,
        'SearchDefinition' => $SearchDefinition,
        'Parameters' => {
            'PortalConfig.ID'  => $PortalConfig->id,
            'SearchLanguageID' => $Servlet->languageID,
        },
        'UriParameters'  => {},
    );
    LogDebug('Options', \%Options);

    my $Search = DE_EPAGES::Search::API::Object::PagerSearch->new(\%Options);
    return $Search->page;
}

#==============================================================================
sub getExportFileName {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    return $PortalConfig->get('Portal')->get('ExportPublicPath') . '/' . $PortalConfig->get('ExportFileName');
}

#==============================================================================
sub ExportPortalConfig {
    my $self = shift;
    my ($Servlet, $DoRedirect) = @_;
    $DoRedirect = 1 unless defined $DoRedirect;

    # save first
    $self->Save($Servlet);

    #now export
    my $PortalConfig = $Servlet->object;
    my $PathFileName = $self->getExportFileName($Servlet);

    my $PortalSite = $PortalConfig->get('PortalSite');
    if ($PortalSite->get('ExportsShippingPrice')) {
        my $ShippingMethod = $PortalConfig->get('ShippingMethod');

        # throw error if shop has no usable shipping methods
        if (!defined($ShippingMethod) ||
            (defined($PortalSite->get('TaxModel')) &&
                !$ShippingMethod->usableAtStorefront({'CurrencyID' => $PortalConfig->get('CurrencyID'), 'TaxModel' => $PortalSite->get('TaxModel')}))
        ) {
            $Servlet->form->addFormError({
                'Form'  => 'ExportPortalConfig',
                'Reason'=> 'NoShipping',
                'Name'  => 'ExportPortalConfig',
                'Value' => 'ShippingMethod'
            }) ;
        }
    }

    my $aWarnings = BlockWarnings( sub {
        eval {
            $PortalConfig->exportPortal($PathFileName);
        };
        if (ExistsError()) {
            my $Error = GetError();
            if ($Error->code eq 'TaxModelMismatch') {
                Warning($Error);
            }
            else {
                Error($Error);
            }
            DeleteFile($PathFileName);
        }
    }, undef, MAX_WARNING_MESSAGES, 0);

    LogDebug('Export Warnings', $aWarnings);
    $Servlet->vars( 'Redirect', 1) if $DoRedirect;
    foreach my $warning (@$aWarnings) {
        if ($warning->code eq 'NativeWarn') {
            Warning($warning);
        }
        else {
            $Servlet->form->mapFormError({
                'Form'  => 'ExportPortalConfig',
                'Name'  => 'ExportPortalConfig'
            }, $warning);
        }
    }
    $Servlet->form->executeFormError;
    $Servlet->vars('ExportSuccessful', 1);

    return;
}

#==============================================================================
sub GeneratePortalConfigProductPrettyUnityUrls {
    my $self = shift;
    my ($Servlet) = @_;

    my $PortalConfig = $Servlet->object();
    my $Shop = $PortalConfig->getSite();

    return unless ($Shop->isUnity());
    return unless ($PortalConfig->isConfigured());

    my $PortalLanguageID = $PortalConfig->get('LanguageID');
    my $aProducts = GetProductPrettyUnityUrlsRequest($PortalConfig->items(), $PortalLanguageID);

    my $ProductsCount = scalar @{ $aProducts };
    return unless ($ProductsCount);

    my $ByteLengthEncodedProducts = $self->_GetJsonByteLength($aProducts);

    if ($ByteLengthEncodedProducts > MAX_REQUEST_BYTE_SIZE) {
        $self->_DoUnityUrlRequestSplitting($Shop, $PortalConfig, $aProducts);
    } else {
        $self->_GetAndSaveProductPrettyUnityUrls($Shop, $PortalConfig, $aProducts);
    }

    return;
}

#==========================================================================================================
# function     _DoUnityUrlRequestSplitting
# state        private
#----------------------------------------------------------------------------------------------------------
# syntax       _DoUnityUrlRequestSplitting( $Shop, $PortalConfig, $Products );
#----------------------------------------------------------------------------------------------------------
# description  If we have too much Products we have to split the content in several requests
#               For calculating the JSON overhead we start the length at 2 accounting for the list chars []
#               We add also 1 char for each element on the list
#----------------------------------------------------------------------------------------------------------
# input        $Shop         | Shop         | DE_EPAGES::Shop::API::Object::Shop
# input        $PortalConfig | PortalConfig | DE_EPAGES::ProductPortal::API::Object::PortalConfig
# input        $Products     | Products     | refArray
#==========================================================================================================
sub _DoUnityUrlRequestSplitting {
    my $self = shift;
    my ($Shop, $PortalConfig, $Products) = @_;

    my @ProductsToSend = ();
    my $ByteLengthEncodedProducts = 2;
    my $ProductByteLength = 0;

    foreach my $hProduct (@{ $Products }) {

        $ProductByteLength = $self->_GetJsonByteLength($hProduct);
        if ($ProductByteLength < MAX_REQUEST_BYTE_SIZE) {
            push(@ProductsToSend, $hProduct);
            $ByteLengthEncodedProducts += $ProductByteLength + 1;
        }

        if ($ByteLengthEncodedProducts > MAX_REQUEST_BYTE_SIZE ) {
            pop(@ProductsToSend);

            $self->_GetAndSaveProductPrettyUnityUrls($Shop, $PortalConfig, \@ProductsToSend);
            @ProductsToSend = ($hProduct);
            $ByteLengthEncodedProducts = $ProductByteLength + 2;
        }
    }

    my $ProductsToSendCount = scalar @ProductsToSend;
    if ($ProductsToSendCount) {
        $self->_GetAndSaveProductPrettyUnityUrls($Shop, $PortalConfig, \@ProductsToSend) ;
    }
    return;
}
#==============================================================================
sub _GetJsonByteLength {
    my $self = shift;
    my ($aList) = @_;

    my $EncodedList = encode('UTF-8',to_json($aList));
    return length($EncodedList);
}

#==============================================================================
sub _GetAndSaveProductPrettyUnityUrls {
    my $self = shift;
    my ($Shop, $PortalConfig, $Products) = @_;

    my $PortalSite = $PortalConfig->get('PortalSite');
    my $PortalLanguageID = $PortalConfig->get('LanguageID');

    my $aResponse = $Shop->sendUnityRequest(
            'POST',
            '/products/get-urls',
            {'locale' => $PortalSite->get('LocaleID')},
            $Products,
            [HTTP_SERVICE_UNAVAILABLE]
    );

    if (defined($aResponse)) {
        SaveProductPrettyUrlsResponse($aResponse, $PortalLanguageID);
    }

    return;
}

#==============================================================================
sub ViewExportResult {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    my $PathFileName = $self->getExportFileName($Servlet);

    my $rContent = GetFileContent($PathFileName, {'binary'=>1});
    my $Response = $Servlet->response;
    $Response->content($$rContent);
    $Response->header(
        'Content-Type'        => 'application/text',
        'Content-Disposition' => 'attachment; filename=' . $PortalConfig->get('Portal')->get('FileName'),
        'Content-Description' => 'portal export file' );

    $Servlet->vars('ExitAfterEvent',1);
    return;
}

#==============================================================================
sub AddProduct{
    my $self = shift;
    my ($Servlet) = @_;

    my $PortalConfig = $Servlet->object;
    my $Form = $Servlet->form;
    my $Alias = $Form->value('NewProduct');
    return unless (defined($Alias) && $Alias =~ /\S/);

    my $TargetProduct = GetProduct($PortalConfig->getSite(), trim($Alias), 1);
    $Form->executeFormError({
        'Reason'=> 'ERROR_INVALIDPRODUCT',
        'Form'  => 'AddProduct',
        'Name'  => 'NewProduct',
    }) unless defined $TargetProduct;

    my $aWarnings = BlockWarnings( sub {
        $PortalConfig->addProducts([$TargetProduct]);
    });
    foreach my $warning (@$aWarnings) {
        $Form->mapFormError({
            'Form'  => 'AddProduct',
            'Name'  => 'NewProduct'
        }, $warning);
    }
    $Form->executeFormError;
    return;
}

#==============================================================================
sub UploadPortalConfig {
    my $self = shift;
    my ($Servlet) = @_;
    my $PortalConfig = $Servlet->object;

    my $UploadSuccessful = $PortalConfig->uploadPortal(); # 1 - success, 0 - failure - -1 - no upload necessary
    if ($UploadSuccessful == 1) {
        $Servlet->vars('UploadSuccessful', 1);
    }
    elsif ($UploadSuccessful == 0) {
        $Servlet->vars('UploadNotSuccessful', 1);
    }

    return;
}

#==============================================================================
sub Save {
    my $self = shift;
    my ($Servlet) = @_;

    my $PortalConfig = $Servlet->object;
    my $Form = $Servlet->form;

    my $Shop    = $PortalConfig->getSite(),
    my $hValues = $Form->form($PortalConfig, 'PortalConfigSettings', 1);

    my $PortalSite = $PortalConfig->get('PortalSite');
    if ($PortalSite->get('ExportsShippingPrice')) {
        my $ShippingMethod = $hValues->{'ShippingMethod'};
        if (!isValid($ShippingMethod)) {
            $Form->addFormError({
                'Form'   => 'PortalConfigSettings',
                'Reason' => 'NoShipping',
                'Name'   => 'ShippingMethod'
            });
        }
        else {
            $ShippingMethod = $Shop->loadObject($ShippingMethod, 'ShippingMethod');
            my $hPortalConfig = {
                'CurrencyID' => $PortalConfig->get('CurrencyID'),
                'TaxModel'   => $PortalSite->get('TaxModel')
            };

            # throw error if shipping method has no price
            $Servlet->form->addFormError({
                'Form'   => 'PortalConfigSettings',
                'Reason' => 'NoShipping',
                'Name'   => 'ShippingMethod'
            }) unless ($ShippingMethod->canAddToBasket($hPortalConfig));
        }
    }
    $Form->executeFormError;

    return $self->SUPER::Save($Servlet);
}

1;

