#========================================================================================
# package      DE_EPAGES::ServiRed::API::Object::PaymentMethodServiRed
#----------------------------------------------------------------------------------------
# state        public
#----------------------------------------------------------------------------------------
# base         DE_EPAGES::Order::API::Object::ContainerPaymentMethod
#----------------------------------------------------------------------------------------
# description  object interface for Payment.
#========================================================================================
package DE_EPAGES::ServiRed::API::Object::PaymentMethodServiRed;
use base DE_EPAGES::Order::API::Object::ContainerPaymentMethod;

use strict;

#========================================================================================
# function     featureName
# state        public
#----------------------------------------------------------------------------------------
# syntax       $FeatureName = $PaymentMethodServiRed->featureName;
#----------------------------------------------------------------------------------------
# description  Returns the feature name 'ServiRed'.
#----------------------------------------------------------------------------------------
# return       $FeatureName | feature name (undef means no feature count) | String
#========================================================================================
sub featureName () { return 'ServiRed'; }

#========================================================================================
# function     usableAtStorefront
# state        public
#----------------------------------------------------------------------------------------
# syntax       $Usable = $PaymentMethod->usableAtStorefront($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a payment method with the given currency is visible in the storefront.
#               Uses function <function DE_EPAGES::Payment::API::Object::PaymentMethod::price>
#               to evaluate if payment methode has a price (price can be 0).
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $Usable | true if payment method is usable in the storefront | boolean
#========================================================================================
sub usableAtStorefront {
    my $self = shift;
    my ($hVars) = @_;

    $self->get('PaymentType')->get('IsActivated') &&
    defined $self->get('SecretKey') &&
    defined $self->get('MerchantCode');
}

#========================================================================================
# function     default
# state        private
#----------------------------------------------------------------------------------------
# syntax       $hValues = DE_EPAGES::4B::API::Object::PaymentMethod4B->default($Parent, $hInfo);
# example      $hDefaults = $Class->defaultAttributes($Parent, $hInfo); # for insert a new object
#               $hDefaults = $Object->class->defaultAttributes($Object->parent, $hValues); # for an existing object
#----------------------------------------------------------------------------------------
# description  Returns default attribute values for new objects.
#               Used by <function DE_EPAGES::Object::Object::Class::defaultAttributes>
#----------------------------------------------------------------------------------------
# input        $Parent | parent object of the new object | object
# input        $hInfo | attribute name => value hash (passed to insert()) | ref.hash
# return       $hValues | hash with default attribute values (name => value) | ref.hash
#========================================================================================
sub default {
    my $self = shift;
    my ($Parent, $hInfo) = @_;

    return {
        %{$self->SUPER::default($Parent, $hInfo)},
        'IsInTestMode'     => 1
    };
}

#========================================================================================
# function     canAddToBasket
# state        public
#----------------------------------------------------------------------------------------
# syntax       if ($PaymentMethod->canAddToBasket($hVars);
#----------------------------------------------------------------------------------------
# description  Tests if a payment method can be added to the basket with the given currency.
#               Overrides function <function DE_EPAGES::Order::API::Object::ContainerPaymentMethod::canAddToBasket>
#               The payment method can be added to the container if the AccountCurrencyID
#               attribute is not set or matches the currency of the container.
#----------------------------------------------------------------------------------------
# input        $hVars | <ul>
#                   <li>CurrencyID    - alpha currency code (ISO 4217) - char(3)
#                   <li>TaxModel      - gross or net price    - int
#               </ul> | ref.hash
# return       $canAddToBasket | true if payment method can be used for the given currency | boolean
#========================================================================================
sub canAddToBasket {
    my $self = shift;
    my ($hVars) = @_;

    return 0 unless $self->SUPER::canAddToBasket($hVars);

    my $PaymentTypeAlias = $self->get('PaymentType')->get('Alias');

    if($PaymentTypeAlias eq 'Bizum'){
        return $hVars->{'CurrencyID'} eq 'EUR'; # Bizum only supports EUR
    }

    return $hVars->{'CurrencyID'} eq 'EUR' || $hVars->{'CurrencyID'} eq 'USD'; # ServiRed only supports EUR and USD
}

1;
