#========================================================================================
# package      DE_EPAGES::ServiRed::API::ParamsRequest
# base         DE_EPAGES::EPagesJ::API::Object::RestClient
# state        public
# description  Client for RedSys Params requests
#========================================================================================
package DE_EPAGES::ServiRed::API::ParamsRequest;
use base DE_EPAGES::EPagesJ::API::Object::RestClient;

use strict;

use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::EPagesJ::API::Object::UriTemplate;
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::ServiRed::API::Constants qw (
    %SERVIRED_UNENCRYPTED_PARAMETERS
);
use Readonly;

Readonly my %URI_TEMPLATES => (
    RedSysEncrypt => 'payments/redsys/encrypt',
    RedSysDecrypt => 'payments/redsys/decrypt',
);

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       DE_EPAGES::ServiRed::API::ParamsRequest->new( %args );
#----------------------------------------------------------------------------------------
# description  Instanciates a new OAuthTokenClient.
#----------------------------------------------------------------------------------------
# input        %args | Input arguments. Possible keys.<ul>
#                 <li>AppsUriTemplate - </li>
#                 <li>AppUriTemplate - ... - DE_EPAGES::EPagesJ::API::Object::UriTemplate</li>
#                 <li>LogCategory - category for commlogger. Defaults to "AppStore" - string</li>
#               </ul> | hash
# return       OAuthTokenClient | new instance | object
#========================================================================================
sub new {
    my $class = shift;
    my %args  = @_;
    $args{LogCategory} //= 'ServiRed';
    $args{DefaultHeaders} = { 'Content-type' => 'application/x-www-form-urlencoded', 'Accept' => 'application/json' };

    return $class->SUPER::new(%args);
}

#========================================================================================
# function     create
# state        public
#----------------------------------------------------------------------------------------
# syntax       $TokenClient = DE_EPAGES::ServiRed::API::ParamsRequest->create;
#----------------------------------------------------------------------------------------
# description  Factory method for a OAuthTokenClient object.
#               Initializes based on the current configuration.
#----------------------------------------------------------------------------------------
# input        $hArgs | Input arguments. Possible keys.<ul>
#                 <li>BaseURL</li>
#                 <li>ShopType</li>
#                 <li>BusinessUnit</li>
#                 <li>Shop</li> | ref.hash
# return       $TokenClient | OAuthTokenClient object | DE_EPAGES::OAuthTokenClient::API::OAuth::OAuthTokenClient
#========================================================================================
sub create {
    my $class = shift;
    my ($hArgs) = @_;

    my $System = LoadRootObject;
    $hArgs->{BaseURL}      //= $System->get('OAuthServiceBaseUrl') // $System->get('EPagesJBaseURL');
    $hArgs->{ShopType}     //= $hArgs->{Shop}->get('ShopTypeRef')->alias;
    $hArgs->{Shop}         //= $hArgs->{Shop};
    $hArgs->{BusinessUnit} //= $System->get('ProviderID');
    foreach my $key (keys %URI_TEMPLATES) {
        $hArgs->{UriTemplates}->{$key} = DE_EPAGES::EPagesJ::API::Object::UriTemplate->new($hArgs->{BaseURL}, $URI_TEMPLATES{$key});
    }

    return $class->SUPER::create( $System, $hArgs );
}

#========================================================================================
# function     createEncryptRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $paramsRequest->createEncryptRequest($hRequest);
#----------------------------------------------------------------------------------------
# description  It retrieves the crypted parameters from the epagesj endpoint
#----------------------------------------------------------------------------------------
# input        $hRequest | Input arguments. Possible keys.<ul>
#                 <li>merchantAmount</li>
#                 <li>merchantOrder</li>
#                      ...
#                 <li>secretKey</li>
# return       $Response | epagesj endpoint url | object
#========================================================================================
sub createEncryptRequest {
    my $self = shift;
    my ($hRequest) = @_;

    return unless isValid($hRequest);
    $self->_validateEncryptRequest($hRequest);

    my $url = $self->_createEncryptRequestUrl();
    my $FormParams = URI->new();
    $FormParams->query_form(
        'merchantAmount'             => $hRequest->{'DS_MERCHANT_AMOUNT'},
        'merchantOrder'              => $hRequest->{'DS_MERCHANT_ORDER'},
        'merchantCode'               => $hRequest->{'DS_MERCHANT_MERCHANTCODE'},
        'merchantCurrency'           => $hRequest->{'DS_MERCHANT_CURRENCY'},
        'merchantTransactionType'    => $hRequest->{'DS_MERCHANT_TRANSACTIONTYPE'},
        'merchantTerminal'           => $hRequest->{'DS_MERCHANT_TERMINAL'},
        'merchantConsumerLanguage'   => $hRequest->{'DS_MERCHANT_CONSUMERLANGUAGE'},
        'merchantUrl'                => $hRequest->{'DS_MERCHANT_MERCHANTURL'},
        'merchantUrlOk'              => $hRequest->{'DS_MERCHANT_URLOK'},
        'merchantUrlKo'              => $hRequest->{'DS_MERCHANT_URLKO'},
        'merchantProductDescription' => $hRequest->{'DS_MERCHANT_PRODUCTDESCRIPTION'},
        'merchantPayMethods'         => $hRequest->{'DS_MERCHANT_PAYMETHODS'},
        'secretKey'                  => $hRequest->{'secretKey'}
    );

    my $FormParamsString = $FormParams->query();
    my $Response = $self->client->post($url, {}, $FormParamsString);
    return $self->_decodeFromResponse($Response);
}

#========================================================================================
sub _validateEncryptRequest {
    my $self = shift;
    my ($hRequest) = @_;

    foreach my $parameter ( keys %SERVIRED_UNENCRYPTED_PARAMETERS ) {
        if ( !isValid($hRequest->{$parameter}) ){
            Error('Mandatory Field empty');
        }
    }

    return;
}

#========================================================================================
sub _createEncryptRequestUrl {
    my $self = shift;
    return $self->{UriTemplates}->{RedSysEncrypt}->uri( { } );
}

#========================================================================================
sub _decodeFromResponse {
    my $self = shift;
    my ($Response) = @_;

    return $self->decodeResponse( $Response->decoded_content() );
}

#========================================================================================
# function     createDecryptRequest
# state        public
#----------------------------------------------------------------------------------------
# syntax       $paramsRequest->createDecryptRequest($hRequest);
#----------------------------------------------------------------------------------------
# description  It retrieves the Decrypted parameters from the epagesj endpoint
#----------------------------------------------------------------------------------------
# input        $hRequest | Input arguments. Possible keys.<ul>
#                 <li>merchantParameters</li>
#                 <li>secretKey</li>
# return       $Response | epagesj endpoint url | object
#========================================================================================
sub createDecryptRequest {
    my $self = shift;
    my ($hRequest) = @_;

    return unless isValid($hRequest);
    $self->_validateDecryptRequest($hRequest);

    my $url = $self->_createDecryptRequestUrl();
    my $FormParams = URI->new();
    $FormParams->query_form(
        'merchantParameters' => $hRequest->{'merchantParameters'},
        'secretKey'          => $hRequest->{'secretKey'}
    );
    my $FormParamsString = $FormParams->query();

    my $Response = $self->client->post($url, {}, $FormParamsString);

    return $self->_decodeFromResponse($Response);
}

#========================================================================================
sub _validateDecryptRequest {
    my $self = shift;
    my ($hRequest) = @_;

    if ( !isValid($hRequest->{'merchantParameters'}) || !isValid($hRequest->{'secretKey'}) ) {
        Error('Mandatory Field empty');
    }

    return;
}

#========================================================================================
sub _createDecryptRequestUrl {
    my $self = shift;
    return $self->{UriTemplates}->{RedSysDecrypt}->uri( { } );
}

1;
