#========================================================================================
# package      DE_EPAGES::ServiRed::API::PaymentSHA256
# state        public
#----------------------------------------------------------------------------------------
# description  Payment functions for ServiRed version HMAC SHA256
#========================================================================================
package DE_EPAGES::ServiRed::API::PaymentSHA256;
use base qw( Exporter );
use strict;

use DE_EPAGES::Core::API::Error qw ( Error );
use DE_EPAGES::Core::API::PerlTools qw ( isValid );
use DE_EPAGES::Object::API::Table::Currency qw ( GetInfoCurrency );
use DE_EPAGES::Object::API::Table::Language qw ( GetInfoLanguage );
use DE_EPAGES::ServiRed::API::Constants qw ( %SERVIRED_LANGUAGES NO_SERVIRED_LANGUAGE DS_MERCHANT_PAYMETHODS_BIZUM DS_MERCHANT_PAYMETHODS_SERVIRED);
use DE_EPAGES::Shop::API::Url qw ( BuildShopUrl );

use constant MINOR_UNIT_EXPONENTIAL_BASE => 10;
use constant SHA256_KEY_LENGTH => 32;

#========================================================================================
# function     new
# state        public
#----------------------------------------------------------------------------------------
# syntax       $RedSys = DE_EPAGES::ServiRed::API::PaymentSHA256->new($Servlet);
#----------------------------------------------------------------------------------------
# description  Creates a new RedSys HMAC SHA256 object .
#----------------------------------------------------------------------------------------
# input        $Servlet | servlet object | object
# return       $RedSys | redsys object | object
#========================================================================================
sub new {
    my $class = shift;
    my ($Servlet) = @_;

    my $self = bless {}, $class;
    $self->{'Servlet'} = $Servlet;

    return $self;
}

#========================================================================================
# function     getSecretKey
# state        public
#----------------------------------------------------------------------------------------
# syntax       getSecretKey();
#----------------------------------------------------------------------------------------
# description  Get secret key from RedSys payment
#               RedSys
#----------------------------------------------------------------------------------------
# return       $hTransferData | hTransferData | refhash
#========================================================================================
sub getSecretKey {
    my $self = shift;

    my $Servlet = $self->{'Servlet'};
    my $BasketOrder = $Servlet->object();
    my $LineItemPayment = $BasketOrder->container->get('Payment');
    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');

    $self->{'SecretKey'} = $PaymentMethod->get('SecretKey');

    return $self->{'SecretKey'};
}


#========================================================================================
# function     collectData
# state        public
#----------------------------------------------------------------------------------------
# syntax       $hMerchantPlainData = $paymentSha256->collectData();
#----------------------------------------------------------------------------------------
# description  Get plain RedSys/Servired parameters
#----------------------------------------------------------------------------------------
# return       $hMerchantPlainData | ref hash
#========================================================================================
sub collectData {
    my $self = shift;

    my $hServiRedParameters = {};

    my $Servlet = $self->{'Servlet'};
    my $BasketOrder = $Servlet->object();

    my $Container = $BasketOrder->container();
    my $LineItemPayment = $Container->get('Payment');
    my $Shop = $LineItemPayment->getSite();
    my $PaymentMethod = $LineItemPayment->get('PaymentMethod');
    my $ContainerID = $LineItemPayment->container->id;

    # language
    my $code2 = GetInfoLanguage($Servlet->languageID())->{'Code2'};

    # currency, amount
    my $hCurrency = GetInfoCurrency($LineItemPayment->get('CurrencyID'));

    # product description

    my $productDescription = 'Shop: ' . $Shop->alias() . ' ;';

    my $merchantMail = $Shop->get('MailFrom');
    if (isValid($merchantMail)) {
        $productDescription .= ' - Contact: ' . $merchantMail;
    }

    #Pay methods (servired or bizum)
    my $payMethodsValue = _GetParameterPayMethodsValue($LineItemPayment);

    # fill hash
    $hServiRedParameters->{'DS_MERCHANT_AMOUNT'} = sprintf('%0.0f', $LineItemPayment->get('Amount') * MINOR_UNIT_EXPONENTIAL_BASE ** $hCurrency->{'MinorUnit'});
    $hServiRedParameters->{'DS_MERCHANT_ORDER'} = $LineItemPayment->get('TransID');
    $hServiRedParameters->{'DS_MERCHANT_MERCHANTCODE'} = $PaymentMethod->get('MerchantCode');
    $hServiRedParameters->{'DS_MERCHANT_CURRENCY'} = $hCurrency->{'NumCode'};
    $hServiRedParameters->{'DS_MERCHANT_TRANSACTIONTYPE'} = $LineItemPayment->get('TransactionType')  ?  1 : 0;
    $hServiRedParameters->{'DS_MERCHANT_TERMINAL'} = $PaymentMethod->get('MerchantTerminal');
    $hServiRedParameters->{'DS_MERCHANT_CONSUMERLANGUAGE'} = exists($SERVIRED_LANGUAGES{$code2})  ?  $SERVIRED_LANGUAGES{$code2}  :  NO_SERVIRED_LANGUAGE;
    $hServiRedParameters->{'DS_MERCHANT_MERCHANTURL'} =
        BuildShopUrl($Shop, {'ChangeAction' => 'ServiRedReturn', 'LineItemContainerID' => $ContainerID}, {'UseSSL' => 1}, $Servlet);
    $hServiRedParameters->{'DS_MERCHANT_URLOK'} =
        BuildShopUrl($Shop, {'ChangeAction' => 'CallbackServiRedReturn', 'LineItemContainerID' => $ContainerID}, {'UseSSL' => 1}, $Servlet);
    $hServiRedParameters->{'DS_MERCHANT_URLKO'} =
        BuildShopUrl($Shop, {'ChangeAction' => 'CallbackServiRedCancel', 'LineItemContainerID' => $ContainerID}, {'UseSSL' => 1}, $Servlet);
    $hServiRedParameters->{'DS_MERCHANT_PRODUCTDESCRIPTION'} = $productDescription;
    $hServiRedParameters->{'DS_MERCHANT_PAYMETHODS'} = $payMethodsValue;

    return $hServiRedParameters;
}

#========================================================================================
# function     isSha256Key
# state        public
#----------------------------------------------------------------------------------------
# syntax       $isSha256Key = PaymentSHA256->isSha256Key();
#----------------------------------------------------------------------------------------
# description  It checks if MBO secret key field is ready for new HMAC-SHA256 or
#               old SHA1 flow
#----------------------------------------------------------------------------------------
# return        $isSha256Key | boolean
#========================================================================================
sub isSha256Key {
    my $self = shift;

    if (isValid($self->{'SecretKey'})) {
        if (length($self->{'SecretKey'}) >= SHA256_KEY_LENGTH) {
            return 1;
        }
        return 0;
    }
    Error( 'InvalidSecretKey');
}

#========================================================================================
# function     _GetParameterPayMethodsValue
# state        private
#----------------------------------------------------------------------------------------
# syntax       my $payMethodsValue = _GetParameterPayMethodsValue($LineItemPayment);
#----------------------------------------------------------------------------------------
# description  Returns the value for the "PayMethods" param  depending on the 
#               PaymentType ('z' for Bizum, 'C' for ServiRed)
#----------------------------------------------------------------------------------------
# input        $LineItemPayment | PaymentType object | object
# return       $payMethodsValue | string
#========================================================================================
sub _GetParameterPayMethodsValue {
    my $LineItemPayment = shift;
    my $PaymentType = $LineItemPayment->get('PaymentType');
    my $isBizum = $PaymentType->get('Alias') eq 'Bizum';
    my $payMethodsValue;
    if($isBizum) {
        $payMethodsValue = DS_MERCHANT_PAYMETHODS_BIZUM;
    } else {
        $payMethodsValue = DS_MERCHANT_PAYMETHODS_SERVIRED;
    }
    return $payMethodsValue;
}

1;