#========================================================================================
# package      DE_EPAGES::Shipping::API::TLE::AvailableShippingFulfillmentTypesHandler
# state        public
#----------------------------------------------------------------------------------------
# description  Implements a TLE Function which returns a list of the available system
#               ShippingTypes
#========================================================================================
package DE_EPAGES::Shipping::API::TLE::AvailableShippingFulfillmentTypesHandler;

use strict;
use DE_EPAGES::Core::API::Log qw ( GetLog );
use DE_EPAGES::Object::API::Factory qw ( LoadRootObject );
use DE_EPAGES::Shipping::API::Tools qw ( GetShopFromTLE IsFeatureAvailable );

#========================================================================================
# function     RegisterHandlerProc
# state        private
#----------------------------------------------------------------------------------------
# syntax       $RegisterHandlerProc( $hParams );
#----------------------------------------------------------------------------------------
# description  this function registers this attribute handle on the processor via hook
#               function on hook TLEProcessorRegistration
#----------------------------------------------------------------------------------------
# input        $hParams | hook arguments keys :
#               <ul>
#                   <li>Processor - tle processor - object
#               </ul> | ref.hash
#========================================================================================
sub RegisterHandlerProc {
    my ($Params) = @_;
    GetLog->debug( 'AvailableShippingFulfillmentTypesHandler.RegisterHandlerProc' );

    __PACKAGE__->new()->register( $Params->{'Processor'} );
    return;
}

#========================================================================================
# function     new
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler = new DE_EPAGES::Shipping::API::TLE::AvailableShippingFulfillmentTypesHandler;
# example      my $PrefixHandler  = new DE_EPAGES::Shipping::API::TLE::AvailableShippingFulfillmentTypesHandler;
#----------------------------------------------------------------------------------------
# description  create a new handler
#----------------------------------------------------------------------------------------
# return       $Handler | blessed object | object
#========================================================================================
sub new { bless {
    'LogisticVendors' => []
    }, shift;
}

#========================================================================================
# function     register
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Handler->register( $Processor );
#----------------------------------------------------------------------------------------
# description  this function register this handle on the processor
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
#========================================================================================
sub register {
    my $self = shift;
    my ($Processor) = @_;
    GetLog->debug( 'AvailableShippingFulfillmentTypesHandler.register' );

    $Processor->registerHandler('VariableHandler', $self, 'AvailableShippingFulfillmentTypes');
    return;
}

#========================================================================================
# function     tle
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->tle( $Processor, $TLEName);
# example      my $Value = $Handler->tle( $Processor, 'Name');
#----------------------------------------------------------------------------------------
# description  This function is called by tle processor to get the value of a tle
#               variable.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | return value | integer
#========================================================================================
sub tle {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    if ( $TLEName eq 'AvailableShippingFulfillmentTypes' ) {
        return $self->_getAvailableShippingFulfillmentTypes( GetShopFromTLE( $Processor ) ) ;
    }

    return undef;
}

#========================================================================================
# function     existsTLE
# state        private
#----------------------------------------------------------------------------------------
# syntax       $Value = $Handler->existsTLE( $Processor, $TLEName);
# example      my $Value = $Handler->existsTLE( $Processor, 'Products_Count');
#----------------------------------------------------------------------------------------
# description  This function is called by the tle processor to test if a TLE variable
#               is available.
#----------------------------------------------------------------------------------------
# input        $Processor | tle processor | object
# input        $TLEName | tle name | string
# return       $Value | yes or no | bool
#========================================================================================
sub existsTLE {
    my $self = shift;
    my ($Processor, $TLEName) = @_;

    return (defined $self->tle($Processor, $TLEName)) ? 1 : 0;
}


#========================================================================================
# function     _getAvailableShippingFulfillmentTypes
# state        private
#----------------------------------------------------------------------------------------
# syntax       _getAvailableShippingFulfillmentTypes($Shop)
# example      $AvailableShippingFulfillmentTypesHandler->_getAvailableShippingFulfillmentTypes($Shop);
#----------------------------------------------------------------------------------------
# description  Get Available ShippingFulfillmentTypes in the shop
#----------------------------------------------------------------------------------------
# input        $Shop | The Shop object | object
# return       $ShippingFulfillmentTypes | array with the available ShippingFulfillmentType | aObject
#========================================================================================
sub _getAvailableShippingFulfillmentTypes {

    my $self = shift;
    my ($Shop) = @_;

    return (defined $Shop) ? $self->_getAvailableLogisticVendors($Shop) : [];
}

#----------------------------------------------------------------------------------------
sub _getAvailableLogisticVendors {
    my $self = shift;

    my ($Shop) = @_;
    $self->_cleanLogisticVendors();

    my $ShippingFulfillmentTypesList = $self->_LoadRootObject()->get('ShippingFulfillmentTypes') ;

    foreach my $ShippingFulfillmentType ( @$ShippingFulfillmentTypesList ) {

        my $ShippingProviderTypes = $ShippingFulfillmentType->get('ShippingProviderTypes');

        my $Feature;

        if (scalar @$ShippingProviderTypes){
            $Feature = $self->_IsFeatureAvailable( $Shop, $ShippingFulfillmentType->get('ProviderMethodClass') ) ;
        } else {
            $Feature = $self->_IsFeatureAvailable( $Shop, $ShippingFulfillmentType->get('MethodClass') ) ;
        }
        if ( $Feature ne 'NotVisible' && !$ShippingFulfillmentType->isDeprecated($Shop) ) {
            $self->_getShippingProviderTypesByShippingFulfillmentType($ShippingFulfillmentType, $Shop);
        }
    }

    return $self->{'LogisticVendors'};
}

#----------------------------------------------------------------------------------------
sub _getShippingProviderTypesByShippingFulfillmentType{
    my $self = shift;
    my ($ShippingFulfillmentType, $Shop) = @_;

    my $ShippingFulfillmentMethodID = $self->_fetchShippingObjectMethodId($Shop->get('ShippingFulfillmentMethods'), $ShippingFulfillmentType);

    my $ShippingProviderMethodID = $self->_fetchShippingObjectMethodId($Shop->get('ShippingProviderMethods'), $ShippingFulfillmentType);

    $self->_addShippingObjectTypes($ShippingFulfillmentType, $ShippingFulfillmentMethodID, $ShippingProviderMethodID);

    return;
}

#----------------------------------------------------------------------------------------
sub _fetchShippingObjectMethodId {
    my $self = shift;
    my ($ShippingMethodsList, $ShippingFulfillmentType) = @_;

    foreach my $ShippingFulfillmentMethod ( @$ShippingMethodsList ) {

        my  $ShippingFulfillmentTypeInUse = $ShippingFulfillmentMethod->get( 'ShippingFulfillmentType')->id ;

        if ( $ShippingFulfillmentType->id eq $ShippingFulfillmentTypeInUse ) {
            return $ShippingFulfillmentMethod->id ;
        }
    }
    return;
}

#----------------------------------------------------------------------------------------
sub _addShippingObjectTypes{
    my $self = shift;
    my ($ShippingFulfillmentType, $ShippingFulfillmentMethodID, $ShippingProviderMethodID) = @_;

    my $ShippingProviderTypes = $ShippingFulfillmentType->get('ShippingProviderTypes');

    if (scalar @$ShippingProviderTypes){
        $self->_addShippingProviderTypes($ShippingProviderTypes, $ShippingFulfillmentType, $ShippingProviderMethodID);
    } else {
        $self->_pushLogisticVendorsArray($ShippingFulfillmentType->tleHash, $ShippingFulfillmentType->tleHash, $ShippingFulfillmentMethodID);
    }
    return;
}

#----------------------------------------------------------------------------------------
sub _addShippingProviderTypes {
    my $self = shift;
    my ($ShippingProviderTypes, $ShippingFulfillmentType, $ShippingProviderMethodID) = @_;

    foreach my $ShippingProviderType ( @$ShippingProviderTypes ) {
        if($ShippingProviderType->get('IsVisible')){
            $self->_pushLogisticVendorsArray($ShippingProviderType->tleHash, $ShippingFulfillmentType->tleHash, $ShippingProviderMethodID);
        }
    }
    return;
}

#----------------------------------------------------------------------------------------
sub _pushLogisticVendorsArray {
    my $self = shift;
    my ($ShippingProviderType, $ShippingFulfillmentType, $ShippingFulfillmentMethodID) = @_;

    push(
        @{$self->{'LogisticVendors'}},
        {
            ShippingProviderType => $ShippingProviderType,
            ShippingFulfillmentType => $ShippingFulfillmentType,
            ShippingFulfillmentMethodID => $ShippingFulfillmentMethodID
        }
    );
    return;
}

#----------------------------------------------------------------------------------------
sub _cleanLogisticVendors {
    my $self = shift;
    $self->{'LogisticVendors'} = [];
    return;
}

#----------------------------------------------------------------------------------------
sub _LoadRootObject {
    my $self = shift;
    return LoadRootObject();
}

#----------------------------------------------------------------------------------------
sub _IsFeatureAvailable {
    my $self = shift;
    my ($Shop, $MethodClass) = @_;
    return IsFeatureAvailable($Shop, $MethodClass);
}

1;
